import enum
import io
import json
import typing as ta


class JsonRenderer:
    class State(enum.Enum):
        VALUE = enum.auto()
        KEY = enum.auto()

    def __init__(
            self,
            out: ta.TextIO,
            *,
            indent: int | str | None = None,
            separators: tuple[str, str] | None = None,
            sort_keys: bool = False,
            style: ta.Callable[[ta.Any, State], tuple[str, str]] | None = None,
    ) -> None:
        super().__init__()

        self._out = out
        if isinstance(indent, (str, int)):
            self._indent = (' ' * indent) if isinstance(indent, int) else indent
            self._endl = '\n'
            if separators is None:
                separators = (',', ': ')
        elif indent is None:
            self._indent = self._endl = ''
            if separators is None:
                separators = (', ', ': ')
        else:
            raise TypeError(indent)
        self._comma, self._colon = separators
        self._sort_keys = sort_keys
        self._style = style

        self._level = 0

    _literals: ta.ClassVar[ta.Mapping[ta.Any, str]] = {
        True: 'true',
        False: 'false',
        None: 'null',
    }

    def _write(self, s: str) -> None:
        if s:
            self._out.write(s)

    def _write_indent(self) -> None:
        if self._indent:
            self._write(self._endl)
            if self._level:
                self._write(self._indent * self._level)

    def _render(self, o: ta.Any, state: State = State.VALUE) -> None:
        if self._style is not None:
            pre, post = self._style(o, state)
            self._write(pre)
        else:
            post = None

        if o is None or isinstance(o, bool):
            self._write(self._literals[o])

        elif isinstance(o, (str, int, float)):
            self._write(json.dumps(o))

        elif isinstance(o, ta.Mapping):
            self._write('{')
            self._level += 1
            items = list(o.items())
            if self._sort_keys:
                items.sort(key=lambda t: t[0])
            for i, (k, v) in enumerate(items):
                if i:
                    self._write(self._comma)
                self._write_indent()
                self._render(k, JsonRenderer.State.KEY)
                self._write(self._colon)
                self._render(v)
            self._level -= 1
            if o:
                self._write_indent()
            self._write('}')

        elif isinstance(o, ta.Sequence):
            self._write('[')
            self._level += 1
            for i, e in enumerate(o):
                if i:
                    self._write(self._comma)
                self._write_indent()
                self._render(e)
            self._level -= 1
            if o:
                self._write_indent()
            self._write(']')

        else:
            raise TypeError(o)

        if post:
            self._write(post)

    def render(self, o: ta.Any) -> None:
        self._render(o)

    @classmethod
    def render_str(cls, o: ta.Any, **kwargs: ta.Any) -> str:
        out = io.StringIO()
        cls(out, **kwargs).render(o)
        return out.getvalue()



import dataclasses as dc
import itertools


MISSING = dc.MISSING


def _dataclass_getstate(self):
    return [getattr(self, f.name) for f in dc.fields(self)]


def _dataclass_setstate(self, state):
    for field, value in zip(dc.fields(self), state):
        object.__setattr__(self, field.name, value)


def _get_slots(cls):
    match cls.__dict__.get('__slots__'):
        # A class which does not define __slots__ at all is equivalent to a class defining
        #   __slots__ = ('__dict__', '__weakref__')
        # `__dictoffset__` and `__weakrefoffset__` can tell us whether the base type has dict/weakref slots, in a way
        # that works correctly for both Python classes and C extension types. Extension types don't use `__slots__` for
        # slot creation
        case None:
            slots = []
            if getattr(cls, '__weakrefoffset__', -1) != 0:
                slots.append('__weakref__')
            if getattr(cls, '__dictrefoffset__', -1) != 0:
                slots.append('__dict__')
            yield from slots
        case str(slot):
            yield slot
        # Slots may be any iterable, but we cannot handle an iterator because it will already be (partially) consumed.
        case iterable if not hasattr(iterable, '__next__'):
            yield from iterable
        case _:
            raise TypeError(f"Slots of '{cls.__name__}' cannot be determined")


def add_slots(
        cls: type,
        is_frozen: bool,
        weakref_slot: bool,
) -> type:
    if '__slots__' in cls.__dict__:
        raise TypeError(f'{cls.__name__} already specifies __slots__')

    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in dc.fields(cls))  # noqa

    inherited_slots = set(itertools.chain.from_iterable(map(_get_slots, cls.__mro__[1:-1])))

    cls_dict['__slots__'] = tuple(
        itertools.filterfalse(
            inherited_slots.__contains__,
            itertools.chain(
                field_names,
                ('__weakref__',) if weakref_slot else (),
            ),
        ),
    )

    for field_name in field_names:
        cls_dict.pop(field_name, None)

    cls_dict.pop('__dict__', None)
    cls_dict.pop('__weakref__', None)

    qualname = getattr(cls, '__qualname__', None)
    cls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        cls.__qualname__ = qualname

    if is_frozen:
        if '__getstate__' not in cls_dict:
            cls.__getstate__ = _dataclass_getstate  # type: ignore
        if '__setstate__' not in cls_dict:
            cls.__setstate__ = _dataclass_setstate  # type: ignore

    return cls



"""
TODO:
 - dogpile
 - midpoint/arc
 - heapq/etc powered lfu
 - thread scoped?
 - __sizeof__ interop
 - further cythonize hot path?
"""
import collections
import contextlib
import functools
import logging
import time
import typing as ta
import weakref

from ... import lang
from .types import Cache
from .types import Eviction
from .types import OverweightError


K = ta.TypeVar('K')
K2 = ta.TypeVar('K2')
V = ta.TypeVar('V')
V2 = ta.TypeVar('V2')


log = logging.getLogger(__name__)


class SKIP(lang.Marker):
    pass


def LRU(cache: 'Cache') -> None:  # noqa
    cache._kill(cache._root.lru_next)  # type: ignore  # noqa


def LRI(cache: 'Cache') -> None:  # noqa
    cache._kill(cache._root.ins_next)  # type: ignore  # noqa


def LFU(cache: 'Cache') -> None:  # noqa
    cache._kill(cache._root.lfu_prev)  # type: ignore  # noqa


class CacheImpl(Cache[K, V]):
    """
    https://google.github.io/guava/releases/16.0/api/docs/com/google/common/cache/CacheBuilder.html
    """

    try:
        if not ta.TYPE_CHECKING:
            from ..._ext.cy.collections.cache import CacheLink as Link
        else:
            raise ImportError  # noqa

    except ImportError:
        class Link:
            __slots__ = [
                'seq',
                'ins_prev',
                'ins_next',
                'lru_prev',
                'lru_next',
                'lfu_prev',
                'lfu_next',
                'key',
                'value',
                'weight',
                'written',
                'accessed',
                'hits',
                'unlinked',
            ]

            seq: int
            ins_prev: 'CacheImpl.Link'
            ins_next: 'CacheImpl.Link'
            lru_prev: 'CacheImpl.Link'
            lru_next: 'CacheImpl.Link'
            lfu_prev: 'CacheImpl.Link'
            lfu_next: 'CacheImpl.Link'
            key: ta.Any | weakref.ref
            value: ta.Any | weakref.ref
            weight: float
            written: float
            accessed: float
            hits: int
            unlinked: bool

            def __repr__(self) -> str:
                return (
                    f'Link@{self.seq!s}('
                    f'ins_prev={("@" + str(self.ins_prev.seq)) if self.ins_prev is not None else None}, '
                    f'ins_next={("@" + str(self.ins_next.seq)) if self.ins_next is not None else None}, '
                    f'lru_prev={("@" + str(self.lru_prev.seq)) if self.lru_prev is not None else None}, '
                    f'lru_next={("@" + str(self.lru_next.seq)) if self.lru_next is not None else None}, '
                    f'lfu_prev={("@" + str(self.lfu_prev.seq)) if self.lfu_prev is not None else None}, '
                    f'lfu_next={("@" + str(self.lfu_next.seq)) if self.lfu_next is not None else None}, '
                    f'key={self.key!r}, '
                    f'value={self.value!r}, '
                    f'weight={self.weight}, '
                    f'written={self.written}, '
                    f'accessed={self.accessed}, '
                    f'hits={self.hits}, '
                    f'unlinked={self.unlinked})'
                )

    _cache: ta.MutableMapping[ta.Any, Link]

    DEFAULT_MAX_SIZE = 256

    def __init__(
            self,
            *,
            max_size: int = DEFAULT_MAX_SIZE,
            max_weight: float | None = None,
            identity_keys: bool = False,
            expire_after_access: float | None = None,
            expire_after_write: float | None = None,
            removal_listener: ta.Callable[[K | weakref.ref, V | weakref.ref], None] | None = None,
            clock: ta.Callable[[], float] | None = None,
            weak_keys: bool = False,
            weak_values: bool = False,
            weigher: ta.Callable[[V], float] = lambda _: 1.,
            lock: lang.DefaultLockable = None,
            raise_overweight: bool = False,
            eviction: Eviction = LRU,
            track_frequency: bool | None = None,
    ) -> None:
        super().__init__()

        if clock is None:
            if expire_after_access is not None or expire_after_write is not None:
                clock = time.time
            else:
                clock = lambda: 0.  # noqa

        self._max_size = max_size
        self._max_weight = max_weight
        self._identity_keys = identity_keys
        self._expire_after_access = expire_after_access
        self._expire_after_write = expire_after_write
        self._removal_listener = removal_listener
        self._clock = clock
        self._weak_keys = weak_keys
        self._weak_values = weak_values
        self._weigher = weigher
        self._lock = lang.default_lock(lock, True)
        self._raise_overweight = raise_overweight
        self._eviction = eviction
        self._track_frequency = track_frequency if track_frequency is not None else (eviction is LFU)

        if weak_keys and not identity_keys:
            self._cache = weakref.WeakKeyDictionary()
        else:
            self._cache = {}

        self._root = CacheImpl.Link()
        self._root.seq = 0
        self._root.ins_next = self._root.ins_prev = self._root
        self._root.lru_next = self._root.lru_prev = self._root
        if self._track_frequency:
            self._root.lfu_next = self._root.lfu_prev = self._root

        weak_dead: collections.deque[CacheImpl.Link] | None
        if weak_keys or weak_values:
            weak_dead = collections.deque()
            weak_dead_ref = weakref.ref(weak_dead)
        else:
            weak_dead = None
            weak_dead_ref = None
        self._weak_dead = weak_dead
        self._weak_dead_ref = weak_dead_ref

        self._seq = 0
        self._size = 0
        self._weight = 0.
        self._hits = 0
        self._misses = 0
        self._max_size_ever = 0
        self._max_weight_ever = 0.

    def _unlink(self, link: Link) -> None:
        if link is self._root:
            raise TypeError
        if link.unlinked:
            return

        link.ins_prev.ins_next = link.ins_next
        link.ins_next.ins_prev = link.ins_prev
        link.ins_next = link.ins_prev = link

        link.lru_prev.lru_next = link.lru_next
        link.lru_next.lru_prev = link.lru_prev
        link.lru_next = link.lru_prev = link

        if self._track_frequency:
            link.lfu_prev.lfu_next = link.lfu_next
            link.lfu_next.lfu_prev = link.lfu_prev
            link.lfu_next = link.lfu_prev = link

        if self._removal_listener is not None:
            try:
                self._removal_listener(link.key, link.value)
            except Exception:
                log.exception('Removal listener raised exception')

        self._size -= 1
        self._weight -= link.weight
        link.key = link.value = None
        link.unlinked = True

    def _kill(self, link: Link) -> None:
        if link is self._root:
            raise RuntimeError

        key = link.key
        if self._weak_keys:
            if key is not None:
                key = key()
            if key is None:
                key = SKIP

        if key is not SKIP:
            cache_key = id(key) if self._identity_keys else key
            cache_link = self._cache.get(cache_key)
            if cache_link is link:
                del self._cache[cache_key]

        self._unlink(link)

    def _reap(self) -> None:
        if self._weak_dead is not None:
            while True:
                try:
                    link = self._weak_dead.popleft()
                except IndexError:
                    break
                self._kill(link)

        clock = None

        if self._expire_after_write is not None:
            clock = self._clock()
            deadline = clock - self._expire_after_write

            while self._root.ins_next is not self._root:
                link = self._root.ins_next
                if link.written > deadline:
                    break
                self._kill(link)

        if self._expire_after_access is not None:
            if clock is None:
                clock = self._clock()
            deadline = clock - self._expire_after_access

            while self._root.lru_next is not self._root:
                link = self._root.lru_next
                if link.accessed > deadline:
                    break
                self._kill(link)

    def reap(self) -> None:
        with self._lock():
            self._reap()

    def _get_link(self, key: K) -> tuple[Link, V]:
        cache_key = id(key) if self._identity_keys else key

        link = self._cache[cache_key]
        if link.unlinked:
            raise Exception

        def fail():
            with contextlib.suppress(KeyError):
                del self._cache[cache_key]
            self._unlink(link)
            raise KeyError(key)

        if self._identity_keys:
            link_key = link.key
            if self._weak_keys:
                link_key = link_key()
                if link_key is None:
                    fail()
            if key is not link_key:
                fail()

        value = link.value
        if self._weak_values:
            if value is not None:
                value = value()
            if value is None:
                fail()

        return link, value  # type: ignore

    def __getitem__(self, key: K) -> V:
        with self._lock():
            self._reap()

            try:
                link, value = self._get_link(key)
            except KeyError:
                self._misses += 1
                raise KeyError(key) from None

            if link.lru_next is not self._root:
                link.lru_prev.lru_next = link.lru_next
                link.lru_next.lru_prev = link.lru_prev

                lru_last = self._root.lru_prev
                lru_last.lru_next = self._root.lru_prev = link
                link.lru_prev = lru_last
                link.lru_next = self._root

            if self._track_frequency:
                lfu_pos = link.lfu_prev
                while lfu_pos is not self._root and lfu_pos.hits <= link.hits:
                    lfu_pos = lfu_pos.lfu_prev

                if link.lfu_prev is not lfu_pos:
                    link.lfu_prev.lfu_next = link.lfu_next
                    link.lfu_next.lfu_prev = link.lfu_prev

                    lfu_last = lfu_pos.lfu_prev
                    lfu_last.lfu_next = lfu_pos.lfu_prev = link
                    link.lfu_prev = lfu_last
                    link.lfu_next = lfu_pos

            link.accessed = self._clock()
            link.hits += 1
            self._hits += 1
            return value

    @staticmethod
    def _weak_die(dead_ref: weakref.ref, link: Link, key_ref: weakref.ref) -> None:  # noqa
        dead = dead_ref()
        if dead is not None:
            dead.append(link)

    @property
    def _full(self) -> bool:
        if self._max_size is not None and self._size >= self._max_size:
            return True
        if self._max_weight is not None and self._weight >= self._max_weight:
            return True
        return False

    def clear(self) -> None:
        with self._lock():
            self._cache.clear()
            while True:
                link = self._root.ins_prev
                if link is self._root:
                    break
                if link.unlinked:
                    raise TypeError
                self._unlink(link)

    def __setitem__(self, key: K, value: V) -> None:
        weight = self._weigher(value)

        with self._lock():
            self._reap()

            if self._max_weight is not None and weight > self._max_weight:
                if self._raise_overweight:
                    raise OverweightError
                else:
                    return

            try:
                existing_link, existing_value = self._get_link(key)
            except KeyError:
                pass
            else:
                self._unlink(existing_link)

            while self._full:
                self._eviction(self)

            link = CacheImpl.Link()

            self._seq += 1
            link.seq = self._seq

            def make_ref(o, b):
                if not b:
                    return o
                return weakref.ref(o, functools.partial(CacheImpl._weak_die, self._weak_dead_ref, link))  # type: ignore  # noqa

            link.key = make_ref(key, self._weak_keys)
            link.value = make_ref(value, self._weak_values)

            link.weight = weight
            link.written = link.accessed = self._clock()
            link.hits = 0
            link.unlinked = False

            ins_last = self._root.ins_prev
            ins_last.ins_next = self._root.ins_prev = link
            link.ins_prev = ins_last
            link.ins_next = self._root

            lru_last = self._root.lru_prev
            lru_last.lru_next = self._root.lru_prev = link
            link.lru_prev = lru_last
            link.lru_next = self._root

            if self._track_frequency:
                lfu_last = self._root.lfu_prev
                lfu_last.lfu_next = self._root.lfu_prev = link
                link.lfu_prev = lfu_last
                link.lfu_next = self._root

            self._weight += weight
            self._size += 1
            self._max_size_ever = max(self._size, self._max_size_ever)
            self._max_weight_ever = max(self._weight, self._max_weight_ever)

            cache_key = id(key) if self._identity_keys else key
            self._cache[cache_key] = link

    def __delitem__(self, key: K) -> None:
        with self._lock():
            self._reap()

            link, value = self._get_link(key)

            cache_key = id(key) if self._identity_keys else key
            del self._cache[cache_key]

            self._unlink(link)

    def __len__(self) -> int:
        with self._lock():
            self._reap()

            return self._size

    def __contains__(self, key: K) -> bool:  # type: ignore
        with self._lock():
            self._reap()

            try:
                self._get_link(key)
            except KeyError:
                return False
            else:
                return True

    def __iter__(self) -> ta.Iterator[K]:
        with self._lock():
            self._reap()

            link = self._root.ins_prev
            while link is not self._root:
                key = link.key
                if self._weak_keys:
                    if key is not None:
                        key = key()
                    if key is not None:
                        yield key
                else:
                    yield key  # type: ignore

                nxt = link.ins_prev
                if nxt is link:
                    raise ValueError
                link = nxt

    @property
    def stats(self) -> Cache.Stats:
        with self._lock():
            return Cache.Stats(
                self._seq,
                self._size,
                self._weight,
                self._hits,
                self._misses,
                self._max_size_ever,
                self._max_weight_ever,
            )


new_cache = CacheImpl



import argparse
import contextlib
import dataclasses as dc
import enum
import json
import subprocess
import sys
import typing as ta

from ... import lang
from ... import term
from .render import JsonRenderer


if ta.TYPE_CHECKING:
    import tomllib

    import yaml

    from .. import dotenv
    from .. import props

else:
    tomllib = lang.proxy_import('tomllib')

    yaml = lang.proxy_import('yaml')

    dotenv = lang.proxy_import('..dotenv', __package__)
    props = lang.proxy_import('..props', __package__)


def term_color(o: ta.Any, state: JsonRenderer.State) -> tuple[str, str]:
    if state is JsonRenderer.State.KEY:
        return term.SGR(term.SGRs.FG.BRIGHT_BLUE), term.SGR(term.SGRs.RESET)
    elif isinstance(o, str):
        return term.SGR(term.SGRs.FG.GREEN), term.SGR(term.SGRs.RESET)
    else:
        return '', ''


@dc.dataclass(frozen=True)
class Format:
    names: ta.Sequence[str]
    load: ta.Callable[[ta.TextIO], ta.Any]


class Formats(enum.Enum):
    JSON = Format(['json'], json.load)
    YAML = Format(['yaml', 'yml'], lambda f: yaml.safe_load(f))
    TOML = Format(['toml'], lambda f: tomllib.loads(f.read()))
    ENV = Format(['env', 'dotenv'], lambda f: dotenv.dotenv_values(stream=f))
    PROPS = Format(['properties', 'props'], lambda f: dict(props.Properties().load(f.read())))


FORMATS_BY_NAME: ta.Mapping[str, Format] = {
    n: f
    for e in Formats
    for f in [e.value]
    for n in f.names
}


def _main() -> None:
    parser = argparse.ArgumentParser()

    parser.add_argument('file', nargs='?')
    parser.add_argument('-f', '--format')
    parser.add_argument('-z', '--compact', action='store_true')
    parser.add_argument('-p', '--pretty', action='store_true')
    parser.add_argument('-i', '--indent')
    parser.add_argument('-s', '--sort-keys', action='store_true')
    parser.add_argument('-c', '--color', action='store_true')
    parser.add_argument('-l', '--less', action='store_true')
    args = parser.parse_args()

    separators = None
    if args.compact:
        separators = (',', ':')

    indent = None
    if args.pretty:
        indent = 2
    if args.indent:
        try:
            indent = int(args.indent)
        except ValueError:
            indent = args.indent

    fmt_name = args.format
    if fmt_name is None:
        if args.file is not None:
            ext = args.file.rpartition('.')[2]
            if ext in FORMATS_BY_NAME:
                fmt_name = ext
    if fmt_name is None:
        fmt_name = 'json'
    fmt = FORMATS_BY_NAME[fmt_name]

    with contextlib.ExitStack() as es:
        if args.file is None:
            in_file = sys.stdin
        else:
            in_file = es.enter_context(open(args.file))

        data = fmt.load(in_file)

    kw: dict[str, ta.Any] = dict(
        indent=indent,
        separators=separators,
        sort_keys=args.sort_keys,
    )

    if args.color:
        out = JsonRenderer.render_str(
            data,
            **kw,
            style=term_color,
        )

    else:
        out = json.dumps(
            data,
            **kw,
        )

    if args.less:
        subprocess.run(
            [
                'less',
                *(['-R'] if args.color else []),
            ],
            input=out.encode(),
            check=True,
        )

    else:
        print(out)


if __name__ == '__main__':
    _main()



import os
import re
import shlex
import sys


##


_DEFAULT_TIMEBOMB_NAME = '-'.join([*__name__.split('.'), 'timebomb'])


def timebomb_payload(delay_s: float, name: str = _DEFAULT_TIMEBOMB_NAME) -> str:
    return (
        '('
        f'echo {shlex.quote(name)} && '
        f'sleep {delay_s:g} && '
        'sh -c \'killall5 -9 -o $PPID -o $$ ; kill 1\''
        ') &'
    )


##


DOCKER_FOR_MAC_HOSTNAME = 'docker.for.mac.localhost'


_LIKELY_IN_DOCKER_PATTERN = re.compile(r'^overlay / .*/docker/')


def is_likely_in_docker() -> bool:
    if getattr(sys, 'platform') != 'linux':
        return False
    with open('/proc/mounts') as f:
        ls = f.readlines()
    return any(_LIKELY_IN_DOCKER_PATTERN.match(l) for l in ls)


##


# Set by pyproject, docker-dev script
DOCKER_HOST_PLATFORM_KEY = 'DOCKER_HOST_PLATFORM'


def get_docker_host_platform() -> str | None:
    return os.environ.get(DOCKER_HOST_PLATFORM_KEY)



"""
TODO:
 - LibreSSL supports aes-256-gcm: https://crypto.stackexchange.com/a/76178
"""
import hashlib
import secrets
import subprocess
import typing as ta

from .. import lang
from .crypto import Crypto
from .crypto import DecryptionError
from .crypto import EncryptionError
from .subprocesses import SubprocessFileInputMethod
from .subprocesses import pipe_fd_subprocess_file_input  # noqa
from .subprocesses import temp_subprocess_file_input  # noqa


if ta.TYPE_CHECKING:
    from cryptography.hazmat.primitives import ciphers as cry_ciphs
    from cryptography.hazmat.primitives.ciphers import algorithms as cry_algs
    from cryptography.hazmat.primitives.ciphers import modes as cry_modes

else:
    cry_ciphs = lang.proxy_import('cryptography.hazmat.primitives.ciphers')
    cry_algs = lang.proxy_import('cryptography.hazmat.primitives.ciphers.algorithms')
    cry_modes = lang.proxy_import('cryptography.hazmat.primitives.ciphers.modes')


##


DEFAULT_KEY_SIZE = 64


def generate_key(self, sz: int = DEFAULT_KEY_SIZE) -> bytes:
    # !! https://docs.openssl.org/3.0/man7/passphrase-encoding/
    # Must not contain null bytes!
    return secrets.token_hex(sz).encode('ascii')


##


class OpensslAescbcCrypto(Crypto):
    """
    !!! https://docs.openssl.org/3.0/man7/passphrase-encoding/
    https://cryptography.io/en/latest/hazmat/primitives/symmetric-encryption/#cryptography.hazmat.primitives.ciphers.Cipher
    https://stackoverflow.com/questions/16761458/how-to-decrypt-openssl-aes-encrypted-files-in-python
    https://github.com/openssl/openssl/blob/3c1713aeed4dc7d1ac25e9e365b8bd98afead638/apps/enc.c#L555-L573
    """

    def __init__(
            self,
            *,
            iters: int = 10_000,
    ) -> None:
        super().__init__()
        self._iters = iters

    def generate_key(self, sz: int = DEFAULT_KEY_SIZE) -> bytes:
        # This actually can handle null bytes, but we don't generate keys with them for compatibility.
        return generate_key(sz)

    block_size = 16
    key_length = 32
    iv_length = 16
    salt_length = 8
    prefix = b'Salted__'

    def _make_cipher(self, key: bytes, salt: bytes) -> 'cry_ciphs.Cipher':
        dk = hashlib.pbkdf2_hmac(
            'sha256',
            key,
            salt,
            self._iters,
            self.key_length + self.iv_length,
        )

        return cry_ciphs.Cipher(
            cry_algs.AES256(dk[:self.key_length]),
            cry_modes.CBC(dk[self.key_length:]),
        )

    def encrypt(self, data: bytes, key: bytes, *, salt: bytes | None = None) -> bytes:
        if salt is None:
            salt = secrets.token_bytes(self.salt_length)
        elif len(salt) != self.salt_length:
            raise EncryptionError('bad salt length')

        last_byte = self.block_size - (len(data) % self.block_size)
        raw = data + bytes([last_byte] * last_byte)

        cipher = self._make_cipher(key, salt)

        encryptor = cipher.encryptor()
        enc = encryptor.update(raw) + encryptor.finalize()

        return b''.join([
            self.prefix,
            salt,
            enc,
        ])

    def decrypt(self, data: bytes, key: bytes) -> bytes:
        if not data.startswith(self.prefix):
            raise DecryptionError('bad prefix')

        salt = data[len(self.prefix):self.block_size]
        cipher = self._make_cipher(key, salt)

        decryptor = cipher.decryptor()
        dec = decryptor.update(data[self.block_size:]) + decryptor.finalize()

        last_byte = dec[-1]
        if last_byte > self.block_size:
            raise DecryptionError('bad padding')

        return dec[:-last_byte]


##


class OpensslSubprocessAescbcCrypto(Crypto):
    def __init__(
            self,
            *,
            cmd: ta.Sequence[str] = ('openssl',),
            timeout: float = 5.,
            file_input: SubprocessFileInputMethod = temp_subprocess_file_input,
    ) -> None:
        super().__init__()
        self._cmd = cmd
        self._timeout = timeout
        self._file_input = file_input

    def generate_key(self, sz: int = DEFAULT_KEY_SIZE) -> bytes:
        return generate_key(sz)

    def encrypt(self, data: bytes, key: bytes) -> bytes:
        # !! https://docs.openssl.org/3.0/man7/passphrase-encoding/
        # Must not contain null bytes!
        if 0 in key:
            raise Exception('invalid key')
        with self._file_input(key) as fi:
            proc = subprocess.Popen(
                [
                    *self._cmd,
                    'aes-256-cbc',

                    '-e',

                    '-pbkdf2',
                    '-salt',
                    '-iter', '10000',

                    '-in', '-',
                    '-out', '-',
                    '-kfile', fi.file_path,
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                pass_fds=[*fi.pass_fds],
            )
            out, err = proc.communicate(
                data,
                timeout=self._timeout,
            )
            if proc.returncode != 0:
                raise EncryptionError
            return out

    def decrypt(self, data: bytes, key: bytes) -> bytes:
        # !! https://docs.openssl.org/3.0/man7/passphrase-encoding/
        # Must not contain null bytes!
        if 0 in key:
            raise Exception('invalid key')
        with self._file_input(key) as fi:
            proc = subprocess.Popen(
                [
                    *self._cmd,
                    'aes-256-cbc',

                    '-d',

                    '-pbkdf2',
                    '-salt',
                    '-iter', '10000',

                    '-in', '-',
                    '-out', '-',
                    '-kfile', fi.file_path,
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                pass_fds=[*fi.pass_fds],
            )
            out, err = proc.communicate(
                data,
                timeout=self._timeout,
            )
            if proc.returncode != 0:
                raise DecryptionError
            return out



import typing as ta
import unicodedata


StrOrBytes: ta.TypeAlias = str | bytes
StrOrBytesT = ta.TypeVar('StrOrBytesT', bound=StrOrBytes)


##


def prefix_delimited(s: StrOrBytesT, p: StrOrBytesT, d: StrOrBytesT) -> StrOrBytesT:
    return d.join([p + l for l in s.split(d)])  # type: ignore


def prefix_lines(s: StrOrBytesT, p: StrOrBytesT) -> StrOrBytesT:
    return prefix_delimited(s, p, '\n' if isinstance(s, str) else b'\n')


def indent_lines(s: StrOrBytesT, num: StrOrBytesT) -> StrOrBytesT:
    return prefix_lines(s, (' ' if isinstance(s, str) else b' ') * num)  # type: ignore


##


def strip_prefix(s: StrOrBytesT, pfx: StrOrBytesT) -> StrOrBytesT:
    if not s.startswith(pfx):  # type: ignore
        raise ValueError(f'{s!r} does not start with {pfx!r}')
    return s[len(pfx):]  # type: ignore


def strip_suffix(s: StrOrBytesT, sfx: StrOrBytesT) -> StrOrBytesT:
    if not s.endswith(sfx):  # type: ignore
        raise ValueError(f'{s!r} does not end with {sfx!r}')
    return s[:-len(sfx)]  # type: ignore


##


def replace_many(
        s: StrOrBytesT,
        old: ta.Iterable[StrOrBytesT],
        new: StrOrBytesT, count_each: int = -1,
) -> StrOrBytesT:
    for o in old:
        s = s.replace(o, new, count_each)  # type: ignore
    return s


##


def camel_case(name: str) -> str:
    return ''.join(map(str.capitalize, name.split('_')))  # noqa


def snake_case(name: str) -> str:
    uppers: list[int | None] = [i for i, c in enumerate(name) if c.isupper()]
    return '_'.join([name[l:r].lower() for l, r in zip([None, *uppers], [*uppers, None])]).strip('_')


def is_dunder(name: str) -> bool:
    return (
        name[:2] == name[-2:] == '__' and
        name[2:3] != '_' and
        name[-3:-2] != '_' and
        len(name) > 4
    )


def is_sunder(name: str) -> bool:
    return (
        name[0] == name[-1] == '_' and
        name[1:2] != '_' and
        name[-2:-1] != '_' and
        len(name) > 2
    )


##


class _CharSpec(ta.NamedTuple):
    cats: ta.AbstractSet[str]
    chars: ta.AbstractSet[str]


_IDENT_START_CHAR_SPEC = _CharSpec(
    {
        'Ll',
        'Lm',
        'Lo',
        'Lt',
        'Lu',
        'Nl',
    },
    {
        '_',
        '\u2118',
        '\u212E',
        '\u309B',
        '\u309C',
    },
)


_IDENT_CONT_CHAR_SPEC = _CharSpec(
    {
        'Ll',
        'Lm',
        'Lo',
        'Lt',
        'Lu',
        'Mc',
        'Mn',
        'Nd',
        'Nl',
        'Pc',
    },
    {
        '_',
        '\u00B7',
        '\u0387',
        '\u1369',
        '\u136A',
        '\u136B',
        '\u136C',
        '\u136D',
        '\u136E',
        '\u136F',
        '\u1370',
        '\u1371',
        '\u19DA',
        '\u2118',
        '\u212E',
        '\u309B',
        '\u309C',
    },
)


def _is_char_in_spec(spec: _CharSpec, c: str) -> bool:
    if len(c) != 1:
        raise ValueError(c)
    return (
        unicodedata.category(c) in spec.cats or
        c in spec.chars or
        unicodedata.category(unicodedata.normalize('NFKC', c)) in spec.cats or
        unicodedata.normalize('NFKC', c) in spec.chars
    )


def is_ident_start(c: str) -> bool:
    return _is_char_in_spec(_IDENT_START_CHAR_SPEC, c)


def is_ident_cont(c: str) -> bool:
    return _is_char_in_spec(_IDENT_CONT_CHAR_SPEC, c)


def is_ident(name: str) -> bool:
    return is_ident_start(name[0]) and all(is_ident_cont(c) for c in name[1:])


##


BOOL_STRINGS: ta.Sequence[tuple[str, str]] = [
    ('n', 'y'),
    ('no', 'yes'),
    ('f', 't'),
    ('false', 'true'),
    ('off', 'on'),
    ('0', '1'),
]

BOOL_FALSE_STRINGS = frozenset(tup[0] for tup in BOOL_STRINGS)
BOOL_TRUE_STRINGS = frozenset(tup[1] for tup in BOOL_STRINGS)

STRING_BOOL_VALUES: ta.Mapping[str, bool] = {
    k: v
    for ks, v in [
        (BOOL_FALSE_STRINGS, False),
        (BOOL_TRUE_STRINGS, True),
    ]
    for k in ks
}



# type: ignore
# ruff: noqa
# flake8: noqa
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#/
from io import StringIO
from typing import Callable
from ..Token import Token
from .ATN import ATN
from .ATNType import ATNType
from .ATNState import *
from .Transition import *
from .LexerAction import *
from .ATNDeserializationOptions import ATNDeserializationOptions

SERIALIZED_VERSION = 4

class ATNDeserializer (object):
    __slots__ = ('deserializationOptions', 'data', 'pos')

    def __init__(self, options : ATNDeserializationOptions = None):
        if options is None:
            options = ATNDeserializationOptions.defaultOptions
        self.deserializationOptions = options

    def deserialize(self, data : [int]):
        self.data = data
        self.pos = 0
        self.checkVersion()
        atn = self.readATN()
        self.readStates(atn)
        self.readRules(atn)
        self.readModes(atn)
        sets = []
        self.readSets(atn, sets)
        self.readEdges(atn, sets)
        self.readDecisions(atn)
        self.readLexerActions(atn)
        self.markPrecedenceDecisions(atn)
        self.verifyATN(atn)
        if self.deserializationOptions.generateRuleBypassTransitions \
                and atn.grammarType == ATNType.PARSER:
            self.generateRuleBypassTransitions(atn)
            # re-verify after modification
            self.verifyATN(atn)
        return atn

    def checkVersion(self):
        version = self.readInt()
        if version != SERIALIZED_VERSION:
            raise Exception("Could not deserialize ATN with version " + str(version) + " (expected " + str(SERIALIZED_VERSION) + ").")

    def readATN(self):
        idx = self.readInt()
        grammarType = ATNType.fromOrdinal(idx)
        maxTokenType = self.readInt()
        return ATN(grammarType, maxTokenType)

    def readStates(self, atn:ATN):
        loopBackStateNumbers = []
        endStateNumbers = []
        nstates = self.readInt()
        for i in range(0, nstates):
            stype = self.readInt()
            # ignore bad type of states
            if stype==ATNState.INVALID_TYPE:
                atn.addState(None)
                continue
            ruleIndex = self.readInt()
            s = self.stateFactory(stype, ruleIndex)
            if stype == ATNState.LOOP_END: # special case
                loopBackStateNumber = self.readInt()
                loopBackStateNumbers.append((s, loopBackStateNumber))
            elif isinstance(s, BlockStartState):
                endStateNumber = self.readInt()
                endStateNumbers.append((s, endStateNumber))

            atn.addState(s)

        # delay the assignment of loop back and end states until we know all the state instances have been initialized
        for pair in loopBackStateNumbers:
            pair[0].loopBackState = atn.states[pair[1]]

        for pair in endStateNumbers:
            pair[0].endState = atn.states[pair[1]]

        numNonGreedyStates = self.readInt()
        for i in range(0, numNonGreedyStates):
            stateNumber = self.readInt()
            atn.states[stateNumber].nonGreedy = True

        numPrecedenceStates = self.readInt()
        for i in range(0, numPrecedenceStates):
            stateNumber = self.readInt()
            atn.states[stateNumber].isPrecedenceRule = True

    def readRules(self, atn:ATN):
        nrules = self.readInt()
        if atn.grammarType == ATNType.LEXER:
            atn.ruleToTokenType = [0] * nrules

        atn.ruleToStartState = [0] * nrules
        for i in range(0, nrules):
            s = self.readInt()
            startState = atn.states[s]
            atn.ruleToStartState[i] = startState
            if atn.grammarType == ATNType.LEXER:
                tokenType = self.readInt()
                atn.ruleToTokenType[i] = tokenType

        atn.ruleToStopState = [0] * nrules
        for state in atn.states:
            if not isinstance(state, RuleStopState):
                continue
            atn.ruleToStopState[state.ruleIndex] = state
            atn.ruleToStartState[state.ruleIndex].stopState = state

    def readModes(self, atn:ATN):
        nmodes = self.readInt()
        for i in range(0, nmodes):
            s = self.readInt()
            atn.modeToStartState.append(atn.states[s])

    def readSets(self, atn:ATN, sets:list):
        m = self.readInt()
        for i in range(0, m):
            iset = IntervalSet()
            sets.append(iset)
            n = self.readInt()
            containsEof = self.readInt()
            if containsEof!=0:
                iset.addOne(-1)
            for j in range(0, n):
                i1 = self.readInt()
                i2 = self.readInt()
                iset.addRange(range(i1, i2 + 1)) # range upper limit is exclusive

    def readEdges(self, atn:ATN, sets:list):
        nedges = self.readInt()
        for i in range(0, nedges):
            src = self.readInt()
            trg = self.readInt()
            ttype = self.readInt()
            arg1 = self.readInt()
            arg2 = self.readInt()
            arg3 = self.readInt()
            trans = self.edgeFactory(atn, ttype, src, trg, arg1, arg2, arg3, sets)
            srcState = atn.states[src]
            srcState.addTransition(trans)

        # edges for rule stop states can be derived, so they aren't serialized
        for state in atn.states:
            for i in range(0, len(state.transitions)):
                t = state.transitions[i]
                if not isinstance(t, RuleTransition):
                    continue
                outermostPrecedenceReturn = -1
                if atn.ruleToStartState[t.target.ruleIndex].isPrecedenceRule:
                    if t.precedence == 0:
                        outermostPrecedenceReturn = t.target.ruleIndex
                trans = EpsilonTransition(t.followState, outermostPrecedenceReturn)
                atn.ruleToStopState[t.target.ruleIndex].addTransition(trans)

        for state in atn.states:
            if isinstance(state, BlockStartState):
                # we need to know the end state to set its start state
                if state.endState is None:
                    raise Exception("IllegalState")
                # block end states can only be associated to a single block start state
                if state.endState.startState is not None:
                    raise Exception("IllegalState")
                state.endState.startState = state

            if isinstance(state, PlusLoopbackState):
                for i in range(0, len(state.transitions)):
                    target = state.transitions[i].target
                    if isinstance(target, PlusBlockStartState):
                        target.loopBackState = state
            elif isinstance(state, StarLoopbackState):
                for i in range(0, len(state.transitions)):
                    target = state.transitions[i].target
                    if isinstance(target, StarLoopEntryState):
                        target.loopBackState = state

    def readDecisions(self, atn:ATN):
        ndecisions = self.readInt()
        for i in range(0, ndecisions):
            s = self.readInt()
            decState = atn.states[s]
            atn.decisionToState.append(decState)
            decState.decision = i

    def readLexerActions(self, atn:ATN):
        if atn.grammarType == ATNType.LEXER:
            count = self.readInt()
            atn.lexerActions = [ None ] * count
            for i in range(0, count):
                actionType = self.readInt()
                data1 = self.readInt()
                data2 = self.readInt()
                lexerAction = self.lexerActionFactory(actionType, data1, data2)
                atn.lexerActions[i] = lexerAction

    def generateRuleBypassTransitions(self, atn:ATN):

        count = len(atn.ruleToStartState)
        atn.ruleToTokenType = [ 0 ] * count
        for i in range(0, count):
            atn.ruleToTokenType[i] = atn.maxTokenType + i + 1

        for i in range(0, count):
            self.generateRuleBypassTransition(atn, i)

    def generateRuleBypassTransition(self, atn:ATN, idx:int):

        bypassStart = BasicBlockStartState()
        bypassStart.ruleIndex = idx
        atn.addState(bypassStart)

        bypassStop = BlockEndState()
        bypassStop.ruleIndex = idx
        atn.addState(bypassStop)

        bypassStart.endState = bypassStop
        atn.defineDecisionState(bypassStart)

        bypassStop.startState = bypassStart

        excludeTransition = None

        if atn.ruleToStartState[idx].isPrecedenceRule:
            # wrap from the beginning of the rule to the StarLoopEntryState
            endState = None
            for state in atn.states:
                if self.stateIsEndStateFor(state, idx):
                    endState = state
                    excludeTransition = state.loopBackState.transitions[0]
                    break

            if excludeTransition is None:
                raise Exception("Couldn't identify final state of the precedence rule prefix section.")

        else:

            endState = atn.ruleToStopState[idx]

        # all non-excluded transitions that currently target end state need to target blockEnd instead
        for state in atn.states:
            for transition in state.transitions:
                if transition == excludeTransition:
                    continue
                if transition.target == endState:
                    transition.target = bypassStop

        # all transitions leaving the rule start state need to leave blockStart instead
        ruleToStartState = atn.ruleToStartState[idx]
        count = len(ruleToStartState.transitions)
        while count > 0:
            bypassStart.addTransition(ruleToStartState.transitions[count-1])
            del ruleToStartState.transitions[-1]

        # link the new states
        atn.ruleToStartState[idx].addTransition(EpsilonTransition(bypassStart))
        bypassStop.addTransition(EpsilonTransition(endState))

        matchState = BasicState()
        atn.addState(matchState)
        matchState.addTransition(AtomTransition(bypassStop, atn.ruleToTokenType[idx]))
        bypassStart.addTransition(EpsilonTransition(matchState))


    def stateIsEndStateFor(self, state:ATNState, idx:int):
        if state.ruleIndex != idx:
            return None
        if not isinstance(state, StarLoopEntryState):
            return None

        maybeLoopEndState = state.transitions[len(state.transitions) - 1].target
        if not isinstance(maybeLoopEndState, LoopEndState):
            return None

        if maybeLoopEndState.epsilonOnlyTransitions and \
                isinstance(maybeLoopEndState.transitions[0].target, RuleStopState):
            return state
        else:
            return None


    #
    # Analyze the {@link StarLoopEntryState} states in the specified ATN to set
    # the {@link StarLoopEntryState#isPrecedenceDecision} field to the
    # correct value.
    #
    # @param atn The ATN.
    #
    def markPrecedenceDecisions(self, atn:ATN):
        for state in atn.states:
            if not isinstance(state, StarLoopEntryState):
                continue

            # We analyze the ATN to determine if this ATN decision state is the
            # decision for the closure block that determines whether a
            # precedence rule should continue or complete.
            #
            if atn.ruleToStartState[state.ruleIndex].isPrecedenceRule:
                maybeLoopEndState = state.transitions[len(state.transitions) - 1].target
                if isinstance(maybeLoopEndState, LoopEndState):
                    if maybeLoopEndState.epsilonOnlyTransitions and \
                            isinstance(maybeLoopEndState.transitions[0].target, RuleStopState):
                        state.isPrecedenceDecision = True

    def verifyATN(self, atn:ATN):
        if not self.deserializationOptions.verifyATN:
            return
        # verify assumptions
        for state in atn.states:
            if state is None:
                continue

            self.checkCondition(state.epsilonOnlyTransitions or len(state.transitions) <= 1)

            if isinstance(state, PlusBlockStartState):
                self.checkCondition(state.loopBackState is not None)

            if isinstance(state, StarLoopEntryState):
                self.checkCondition(state.loopBackState is not None)
                self.checkCondition(len(state.transitions) == 2)

                if isinstance(state.transitions[0].target, StarBlockStartState):
                    self.checkCondition(isinstance(state.transitions[1].target, LoopEndState))
                    self.checkCondition(not state.nonGreedy)
                elif isinstance(state.transitions[0].target, LoopEndState):
                    self.checkCondition(isinstance(state.transitions[1].target, StarBlockStartState))
                    self.checkCondition(state.nonGreedy)
                else:
                    raise Exception("IllegalState")

            if isinstance(state, StarLoopbackState):
                self.checkCondition(len(state.transitions) == 1)
                self.checkCondition(isinstance(state.transitions[0].target, StarLoopEntryState))

            if isinstance(state, LoopEndState):
                self.checkCondition(state.loopBackState is not None)

            if isinstance(state, RuleStartState):
                self.checkCondition(state.stopState is not None)

            if isinstance(state, BlockStartState):
                self.checkCondition(state.endState is not None)

            if isinstance(state, BlockEndState):
                self.checkCondition(state.startState is not None)

            if isinstance(state, DecisionState):
                self.checkCondition(len(state.transitions) <= 1 or state.decision >= 0)
            else:
                self.checkCondition(len(state.transitions) <= 1 or isinstance(state, RuleStopState))

    def checkCondition(self, condition:bool, message=None):
        if not condition:
            if message is None:
                message = "IllegalState"
            raise Exception(message)

    def readInt(self):
        i = self.data[self.pos]
        self.pos += 1
        return i

    edgeFactories = [ lambda args : None,
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : EpsilonTransition(target),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        RangeTransition(target, Token.EOF, arg2) if arg3 != 0 else RangeTransition(target, arg1, arg2),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        RuleTransition(atn.states[arg1], arg2, arg3, target),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        PredicateTransition(target, arg1, arg2, arg3 != 0),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        AtomTransition(target, Token.EOF) if arg3 != 0 else AtomTransition(target, arg1),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        ActionTransition(target, arg1, arg2, arg3 != 0),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        SetTransition(target, sets[arg1]),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        NotSetTransition(target, sets[arg1]),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        WildcardTransition(target),
                      lambda atn, src, trg, arg1, arg2, arg3, sets, target : \
                        PrecedencePredicateTransition(target, arg1)
                      ]

    def edgeFactory(self, atn:ATN, type:int, src:int, trg:int, arg1:int, arg2:int, arg3:int, sets:list):
        target = atn.states[trg]
        if type > len(self.edgeFactories) or self.edgeFactories[type] is None:
            raise Exception("The specified transition type: " + str(type) + " is not valid.")
        else:
            return self.edgeFactories[type](atn, src, trg, arg1, arg2, arg3, sets, target)

    stateFactories = [  lambda : None,
                        lambda : BasicState(),
                        lambda : RuleStartState(),
                        lambda : BasicBlockStartState(),
                        lambda : PlusBlockStartState(),
                        lambda : StarBlockStartState(),
                        lambda : TokensStartState(),
                        lambda : RuleStopState(),
                        lambda : BlockEndState(),
                        lambda : StarLoopbackState(),
                        lambda : StarLoopEntryState(),
                        lambda : PlusLoopbackState(),
                        lambda : LoopEndState()
                    ]

    def stateFactory(self, type:int, ruleIndex:int):
        if type> len(self.stateFactories) or self.stateFactories[type] is None:
            raise Exception("The specified state type " + str(type) + " is not valid.")
        else:
            s = self.stateFactories[type]()
            if s is not None:
                s.ruleIndex = ruleIndex
        return s

    CHANNEL = 0     #The type of a {@link LexerChannelAction} action.
    CUSTOM = 1      #The type of a {@link LexerCustomAction} action.
    MODE = 2        #The type of a {@link LexerModeAction} action.
    MORE = 3        #The type of a {@link LexerMoreAction} action.
    POP_MODE = 4    #The type of a {@link LexerPopModeAction} action.
    PUSH_MODE = 5   #The type of a {@link LexerPushModeAction} action.
    SKIP = 6        #The type of a {@link LexerSkipAction} action.
    TYPE = 7        #The type of a {@link LexerTypeAction} action.

    actionFactories = [ lambda data1, data2: LexerChannelAction(data1),
                        lambda data1, data2: LexerCustomAction(data1, data2),
                        lambda data1, data2: LexerModeAction(data1),
                        lambda data1, data2: LexerMoreAction.INSTANCE,
                        lambda data1, data2: LexerPopModeAction.INSTANCE,
                        lambda data1, data2: LexerPushModeAction(data1),
                        lambda data1, data2: LexerSkipAction.INSTANCE,
                        lambda data1, data2: LexerTypeAction(data1)
                      ]

    def lexerActionFactory(self, type:int, data1:int, data2:int):

        if type > len(self.actionFactories) or self.actionFactories[type] is None:
            raise Exception("The specified lexer action type " + str(type) + " is not valid.")
        else:
            return self.actionFactories[type](data1, data2)



import operator
import typing as ta

from .... import check
from .... import collections as col
from .... import lang
from .base import BooleanKeyword
from .base import Keyword
from .base import Keywords
from .base import KeywordsKeyword
from .base import NumberKeyword
from .base import StrKeyword
from .base import StrOrStrsKeyword
from .base import StrToKeywordsKeyword
from .core import CoreKeyword
from .metadata import MetadataKeyword
from .validation import ValidationKeyword


KeywordT = ta.TypeVar('KeywordT', bound=Keyword)


##


def build_keyword_types_by_tag(keyword_types: ta.Iterable[type[Keyword]]) -> ta.Mapping[str, type[Keyword]]:
    return col.make_map_by(operator.attrgetter('tag'), keyword_types, strict=True)


DEFAULT_KEYWORD_SUPERTYPES: ta.AbstractSet = frozenset([
    CoreKeyword,
    MetadataKeyword,
    ValidationKeyword,
])

DEFAULT_KEYWORD_TYPES: ta.AbstractSet = frozenset(lang.flatten(
    lang.deep_subclasses(st, concrete_only=True) for st in DEFAULT_KEYWORD_SUPERTYPES
))

DEFAULT_KEYWORD_TYPES_BY_TAG: ta.Mapping[str, type[Keyword]] = build_keyword_types_by_tag(DEFAULT_KEYWORD_TYPES)


##


class Parser:
    def __init__(
            self,
            keyword_types: ta.Iterable[type[Keyword]] | ta.Mapping[str, type[Keyword]] = DEFAULT_KEYWORD_TYPES_BY_TAG,
    ) -> None:
        super().__init__()

        if isinstance(keyword_types, ta.Mapping):
            self._keyword_types_by_tag = keyword_types
        else:
            self._keyword_types_by_tag = build_keyword_types_by_tag(keyword_types)

    def parse_keyword(self, cls: type[KeywordT], v: ta.Any) -> KeywordT:
        if issubclass(cls, BooleanKeyword):
            return cls(check.isinstance(v, bool))  # type: ignore

        elif issubclass(cls, NumberKeyword):
            return cls(check.isinstance(v, (int, float)))  # type: ignore

        elif issubclass(cls, StrKeyword):
            return cls(check.isinstance(v, str))  # type: ignore

        elif issubclass(cls, StrOrStrsKeyword):
            ss: str | ta.Sequence[str]
            if isinstance(v, str):
                ss = v
            elif isinstance(v, ta.Iterable):
                ss = col.seq_of(check.of_isinstance(str))(v)
            else:
                raise TypeError(v)
            return cls(ss)  # type: ignore

        elif issubclass(cls, KeywordsKeyword):
            return cls(parse_keywords(v))  # type: ignore

        elif issubclass(cls, StrToKeywordsKeyword):
            return cls({k: parse_keywords(mv) for k, mv in v.items()})  # type: ignore

        else:
            raise TypeError(cls)

    def parse_keywords(self, dct: ta.Mapping[str, ta.Any]) -> Keywords:
        lst: list[Keyword] = []
        for k, v in dct.items():
            cls = self._keyword_types_by_tag[k]
            lst.append(self.parse_keyword(cls, v))
        return Keywords(lst)


##


DEFAULT_PARSER = Parser()

parse_keyword = DEFAULT_PARSER.parse_keyword
parse_keywords = DEFAULT_PARSER.parse_keywords



"""
TODO:
 - parser?
 - js? viz.js, d3, visjs
 - cycle detection
 - networkx adapter
 - https://docs.python.org/3.9/library/graphlib.html#module-graphlib
"""
import typing as ta

from .. import check
from .. import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')
T = ta.TypeVar('T')
U = ta.TypeVar('U')


def traverse_links(data: ta.Mapping[T, ta.Iterable[T]], keys: ta.Iterable[T]) -> set[T]:
    keys = set(keys)
    todo = set(keys)
    seen: set[T] = set()
    while todo:
        key = todo.pop()
        seen.add(key)
        cur = data.get(key, [])
        todo.update(set(cur) - seen)
    return seen - keys


def invert_set_map(src: ta.Mapping[K, ta.Iterable[V]]) -> dict[V, set[K]]:
    dst: dict[V, set[K]] = {}
    for l, rs in src.items():
        for r in rs:
            try:
                s = dst[r]
            except KeyError:
                s = dst[r] = set()
            s.add(l)
    return dst


def invert_symmetric_set_map(src: ta.Mapping[T, ta.Iterable[T]]) -> dict[T, set[T]]:
    dst: dict[T, set[T]] = {l: set() for l in src}
    for l, rs in src.items():
        for r in rs:
            dst[r].add(l)
    return dst


class Dag(ta.Generic[T]):

    def __init__(self, input_its_by_outputs: ta.Mapping[T, ta.Iterable[T]]) -> None:
        super().__init__()

        self._input_sets_by_output = {u: set(d) for u, d in input_its_by_outputs.items()}

    @property
    def input_sets_by_output(self) -> ta.Mapping[T, ta.AbstractSet[T]]:
        return self._input_sets_by_output

    @lang.cached_property
    def output_sets_by_input(self) -> ta.Mapping[T, ta.AbstractSet[T]]:
        return invert_symmetric_set_map(self._input_sets_by_output)

    def subdag(self, *args, **kwargs) -> 'Subdag[T]':
        return Subdag(self, *args, **kwargs)


class Subdag(ta.Generic[U]):

    def __init__(
            self,
            dag: 'Dag[U]',
            targets: ta.Iterable[U],
            *,
            ignored: ta.Iterable[U] | None = None,
    ) -> None:
        super().__init__()

        self._dag: Dag[U] = check.isinstance(dag, Dag)  # type: ignore
        self._targets = set(targets)
        self._ignored = set(ignored or []) - self._targets

    @property
    def dag(self) -> 'Dag[U]':
        return self._dag

    @property
    def targets(self) -> ta.AbstractSet[U]:
        return self._targets

    @property
    def ignored(self) -> ta.AbstractSet[U]:
        return self._ignored

    @lang.cached_property
    def inputs(self) -> ta.AbstractSet[U]:
        return traverse_links(self.dag.input_sets_by_output, self.targets) - self.ignored

    @lang.cached_property
    def outputs(self) -> ta.AbstractSet[U]:
        return traverse_links(self.dag.output_sets_by_input, self.targets) - self.ignored

    @lang.cached_property
    def output_inputs(self) -> ta.AbstractSet[U]:
        return traverse_links(self.dag.input_sets_by_output, self.outputs) - self.ignored

    @lang.cached_property
    def all(self) -> ta.AbstractSet[U]:
        return self.targets | self.inputs | self.outputs | self.output_inputs



import functools


class cached_nullary:  # noqa
    def __init__(self, fn):
        super().__init__()
        self._fn = fn
        self._value = self._missing = object()
        functools.update_wrapper(self, fn)

    def __call__(self, *args, **kwargs):  # noqa
        if self._value is self._missing:
            self._value = self._fn()
        return self._value

    def __get__(self, instance, owner):  # noqa
        bound = instance.__dict__[self._fn.__name__] = self.__class__(self._fn.__get__(instance, owner))
        return bound



"""
TODO:
 - multis?

class SetBinding(Element, lang.Final):
class SetProvider(Provider):

class MapBinding(Element, lang.Final):
class MapProvider(Provider):
"""
import functools
import inspect
import types
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from .bindings import Binding
from .eagers import Eager
from .elements import Element
from .elements import Elements
from .keys import Key
from .keys import as_key
from .privates import Expose
from .providers import ConstProvider
from .providers import CtorProvider
from .providers import FnProvider
from .providers import LinkProvider
from .providers import Provider
from .scopes import SCOPE_ALIASES
from .scopes import Singleton
from .types import Scope
from .types import Unscoped


if ta.TYPE_CHECKING:
    from .impl.inspect import inspect as _inspect
else:
    _inspect = lang.proxy_import('.impl.inspect', __package__)


T = ta.TypeVar('T')


##


_FN_TYPES: tuple[type, ...] = (
    types.FunctionType,
    types.MethodType,

    classmethod,
    staticmethod,

    functools.partial,
    functools.partialmethod,
)


def _is_fn(obj: ta.Any) -> bool:
    return isinstance(obj, _FN_TYPES)


def bind_as_fn(cls: type[T]) -> type[T]:
    check.isinstance(cls, type)
    global _FN_TYPES
    if cls not in _FN_TYPES:
        _FN_TYPES = (*_FN_TYPES, cls)
    return cls


##


_BANNED_BIND_TYPES = (
    Element,
    Provider,
    Elements,
    Scope,
)


def bind(
        obj: ta.Any,
        *,
        tag: ta.Any = None,

        to_fn: ta.Any = None,
        to_ctor: ta.Any = None,
        to_const: ta.Any = None,
        to_key: ta.Any = None,

        in_: Scope | None = None,
        singleton: bool = False,

        eager: bool = False,
        expose: bool = False,
) -> Element | Elements:
    if obj is None or obj is inspect.Parameter.empty:
        raise TypeError(obj)
    if isinstance(obj, _BANNED_BIND_TYPES):
        raise TypeError(obj)

    ##

    has_to = (
        to_fn is not None or
        to_ctor is not None or
        to_const is not None or
        to_key is not None
    )
    if isinstance(obj, Key):
        key = obj
    elif isinstance(obj, type):
        if not has_to:
            to_ctor = obj
        key = Key(obj)
    elif isinstance(obj, rfl.TYPES) or rfl.is_type(obj):
        key = Key(obj)
    elif _is_fn(obj) and not has_to:
        sig = _inspect.signature(obj)
        ty = rfl.type_(sig.return_annotation)
        to_fn = obj
        key = Key(ty)
    else:
        if to_const is not None:
            raise TypeError('Cannot bind instance with to_const')
        to_const = obj
        key = Key(type(obj))
    del has_to

    ##

    if tag is not None:
        if key.tag is not None:
            raise TypeError('Tag already set')
        key = dc.replace(key, tag=tag)

    ##

    providers: list[Provider] = []
    if to_fn is not None:
        providers.append(FnProvider(to_fn))
    if to_ctor is not None:
        providers.append(CtorProvider(to_ctor))
    if to_const is not None:
        providers.append(ConstProvider(to_const))
    if to_key is not None:
        providers.append(LinkProvider(as_key(to_key)))
    if not providers:
        raise TypeError('Must specify provider')
    if len(providers) > 1:
        raise TypeError('May not specify multiple providers')
    provider, = providers

    ##

    scopes: list[Scope] = []
    if in_ is not None:
        if isinstance(in_, str):
            scopes.append(SCOPE_ALIASES[in_])
        else:
            scopes.append(check.isinstance(in_, Scope))
    if singleton:
        scopes.append(Singleton())
    if len(scopes) > 1:
        raise TypeError('May not specify multiple scopes')
    scope: Scope
    if not scopes:
        scope = Unscoped()
    else:
        scope, = scopes

    ##

    binding = Binding(key, provider, scope)

    ##

    elements: list[Element] = [binding]

    if eager:
        elements.append(Eager(key))
    if expose:
        elements.append(Expose(key))

    if len(elements) == 1:
        return elements[0]
    else:
        return Elements(frozenset(elements))



import contextlib
import fcntl
import os
import resource
import shutil
import signal
import tempfile
import typing as ta


PAGE_SIZE = resource.getpagesize()


def round_to_page_size(sz: int) -> int:
    sz += PAGE_SIZE - 1
    return sz - (sz % PAGE_SIZE)


@contextlib.contextmanager
def tmp_dir(
        root_dir: str | None = None,
        cleanup: bool = True,
        **kwargs: ta.Any,
) -> ta.Iterator[str]:
    path = tempfile.mkdtemp(dir=root_dir, **kwargs)
    try:
        yield path
    finally:
        if cleanup:
            shutil.rmtree(path, ignore_errors=True)


@contextlib.contextmanager
def tmp_file(
        root_dir: str | None = None,
        cleanup: bool = True,
        **kwargs: ta.Any,
) -> ta.Iterator[tempfile._TemporaryFileWrapper]:  # noqa
    with tempfile.NamedTemporaryFile(dir=root_dir, delete=False, **kwargs) as f:
        try:
            yield f
        finally:
            if cleanup:
                shutil.rmtree(f.name, ignore_errors=True)


class Pidfile:
    def __init__(self, path: str) -> None:
        super().__init__()
        self._path = path

    _f: ta.TextIO

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._path!r})'

    def __enter__(self) -> ta.Self:
        fd = os.open(self._path, os.O_RDWR | os.O_CREAT, 0o600)
        try:
            os.set_inheritable(fd, True)
            f = os.fdopen(fd, 'r+')
        except Exception:
            try:
                os.close(fd)
            except Exception:  # noqa
                pass
            raise
        self._f = f
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._f is not None:
            self._f.close()
            del self._f

    def try_lock(self) -> bool:
        try:
            fcntl.flock(self._f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True
        except OSError:
            return False

    def write(self, pid: int | None = None) -> None:
        if not self.try_lock():
            raise RuntimeError('Could not get lock')
        if pid is None:
            pid = os.getpid()
        self._f.write(f'{pid}\n')
        self._f.flush()

    def clear(self) -> None:
        if not self.try_lock():
            raise RuntimeError('Could not get lock')
        self._f.seek(0)
        self._f.truncate()

    def read(self) -> int:
        if self.try_lock():
            raise RuntimeError('Got lock')
        self._f.seek(0)
        return int(self._f.read())

    def kill(self, sig: int = signal.SIGTERM) -> None:
        pid = self.read()
        os.kill(pid, sig)  # Still racy


def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
    if exist_ok:
        # First try to bump modification time
        # Implementation note: GNU touch uses the UTIME_NOW option of the utimensat() / futimens() functions.
        try:
            os.utime(self, None)
        except OSError:
            pass
        else:
            return
    flags = os.O_CREAT | os.O_WRONLY
    if not exist_ok:
        flags |= os.O_EXCL
    fd = os.open(self, flags, mode)
    os.close(fd)



import typing as ta

from .. import dataclasses as dc
from .. import lang
from .bindings import Binding
from .elements import Element
from .injector import Injector
from .keys import Key


ProvisionListener: ta.TypeAlias = ta.Callable[[
    Injector,
    Key,
    Binding,
    ta.Callable[[], ta.Any],
], ta.Callable[[], ta.Any]]


@dc.dataclass(frozen=True)
@dc.extra_params(cache_hash=True)
class ProvisionListenerBinding(Element, lang.Final):
    listener: ProvisionListener


def bind_provision_listener(l: ProvisionListener) -> Element:
    return ProvisionListenerBinding(l)



import abc
import types
import typing as ta

from .abstract import make_abstract
from .restrict import Final
from .restrict import NotInstantiable


T = ta.TypeVar('T')
Ty = ta.TypeVar('Ty', bound=type)


##


def _make_not_instantiable():
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError(cls)

    def __init__(self, *args, **kwargs):  # noqa
        raise TypeError(self)

    return {
        '__new__': __new__,
        '__initT__': __init__,
    }


class _VirtualMeta(abc.ABCMeta):

    def __new__(mcls, name, bases, namespace):
        if 'Virtual' not in globals():
            return super().__new__(mcls, name, bases, namespace)
        if Virtual not in bases:
            raise TypeError

        for k, v in list(namespace.items()):
            absv = make_abstract(v)
            if absv is not v:
                namespace[k] = absv

        reqs = {
            k: v
            for k, v in namespace.items()
            if getattr(v, '__isabstractmethod__', False)
        }

        user_subclasshook = namespace.pop('__subclasshook__', None)

        def get_missing_reqs(cls):
            reqset = set(reqs)
            for mro_cls in cls.__mro__:
                reqset -= set(mro_cls.__dict__)
            return reqset

        def __subclasshook__(cls, subclass):  # noqa
            if cls is not kls:
                return super(kls, cls).__subclasshook__(subclass)  # type: ignore
            if get_missing_reqs(subclass):
                return False
            if user_subclasshook is not None:
                ret = user_subclasshook(cls, subclass)
            else:
                ret = super(kls, cls).__subclasshook__(subclass)  # type: ignore
            return True if ret is NotImplemented else ret

        namespace['__subclasshook__'] = classmethod(__subclasshook__)

        namespace.update(_make_not_instantiable())

        kls = ta.cast(
            type,
            super().__new__(
                abc.ABCMeta,
                name,
                tuple(b for b in bases if b is not Virtual),
                namespace,
            ),
        )
        return kls


class Virtual(metaclass=_VirtualMeta):
    """Like Protocol but supports more than just methods."""


def virtual_check(virtual: type) -> ta.Callable[[Ty], Ty]:
    def inner(cls):
        if not issubclass(cls, virtual):
            raise TypeError(cls)
        return cls
    # if not issubclass(type(virtual), _VirtualMeta):
    #     raise TypeError(virtual)
    return inner


##


class Descriptor(Virtual):
    def __get__(self, instance, owner=None):
        raise NotImplementedError


class Picklable(Virtual):
    def __getstate__(self):
        raise NotImplementedError

    def __setstate__(self, state):
        raise NotImplementedError


##


class Callable(NotInstantiable, Final, ta.Generic[T]):
    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        raise TypeError

    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return callable(instance)

    @classmethod
    def __subclasscheck__(cls, subclass: type) -> bool:
        if not hasattr(subclass, '__call__'):  # noqa
            return False
        call = subclass.__call__
        if isinstance(call, types.MethodWrapperType) and call.__self__ is subclass:
            return False
        return True



"""
TODO:
 - -x / --exec - os.exec entrypoint
  - refuse to install non-exec-relevant Bootstraps when chosen
"""
# ruff: noqa: UP006 UP007
import argparse
import dataclasses as dc
import io
import itertools
import sys
import typing as ta

from .. import lang
from .base import Bootstrap
from .harness import BOOTSTRAP_TYPES_BY_NAME
from .harness import bootstrap


if ta.TYPE_CHECKING:
    import runpy

else:
    runpy = lang.proxy_import('runpy')


##


class _OrderedArgsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if 'ordered_args' not in namespace:
            setattr(namespace, 'ordered_args', [])
        if self.const is not None:
            value = self.const
        else:
            value = values
        namespace.ordered_args.append((self.dest, value))


def _or_opt(ty):
    return (ty, ta.Optional[ty])


def _int_or_str(v):
    try:
        return int(v)
    except ValueError:
        return v


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    # ta.Optional[ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]]

    for cname, cls in BOOTSTRAP_TYPES_BY_NAME.items():
        for fld in dc.fields(cls.Config):
            aname = f'--{cname}:{fld.name}'
            kw: ta.Dict[str, ta.Any] = {}

            if fld.type in _or_opt(str):
                pass
            elif fld.type in _or_opt(bool):
                kw.update(const=True, nargs=0)
            elif fld.type in _or_opt(int):
                kw.update(type=int)
            elif fld.type in _or_opt(float):
                kw.update(type=float)
            elif fld.type in _or_opt(ta.Union[int, str]):
                kw.update(type=_int_or_str)

            elif fld.type in (
                    *_or_opt(ta.Sequence[str]),
                    *_or_opt(ta.Mapping[str, ta.Optional[str]]),
                    *_or_opt(ta.Mapping[int, ta.Union[int, str, None]]),
                    *_or_opt(ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]),
            ):
                if aname[-1] != 's':
                    raise NameError(aname)
                aname = aname[:-1]

            else:
                raise TypeError(fld)

            parser.add_argument(aname, action=_OrderedArgsAction, **kw)


def _process_arguments(args: ta.Any) -> ta.Sequence[Bootstrap.Config]:
    if not (oa := getattr(args, 'ordered_args', None)):
        return []

    cfgs: ta.List[Bootstrap.Config] = []

    for cname, cargs in [
        (n, list(g))
        for n, g in
        itertools.groupby(oa, key=lambda s: s[0].partition(':')[0])
    ]:
        ccls = BOOTSTRAP_TYPES_BY_NAME[cname].Config
        flds = {f.name: f for f in dc.fields(ccls)}

        kw: ta.Dict[str, ta.Any] = {}
        for aname, aval in cargs:
            k = aname.partition(':')[2]

            if k not in flds:
                k += 's'
                fld = flds[k]

                if fld.type in _or_opt(ta.Sequence[str]):
                    kw.setdefault(k, []).append(aval)

                elif fld.type in _or_opt(ta.Mapping[str, ta.Optional[str]]):
                    if '=' not in aval:
                        kw.setdefault(k, {})[aval] = None
                    else:
                        ek, _, ev = aval.partition('=')
                        kw.setdefault(k, {})[ek] = ev

                elif fld.type in _or_opt(ta.Mapping[int, ta.Union[int, str, None]]):
                    fk, _, fv = aval.partition('=')
                    if not fv:
                        kw.setdefault(k, {})[int(fk)] = None
                    else:
                        kw.setdefault(k, {})[int(fk)] = _int_or_str(fv)

                elif fld.type in _or_opt(ta.Mapping[str, ta.Tuple[ta.Optional[int], ta.Optional[int]]]):
                    fk, _, fv = aval.partition('=')
                    if ',' in fv:
                        tl, tr = fv.split(',')
                    else:
                        tl, tr = None, None
                    kw.setdefault(k, {})[fk] = (_int_or_str(tl) if tl else None, _int_or_str(tr) if tr else None)

                else:
                    raise TypeError(fld)

            else:
                kw[k] = aval

        cfg = ccls(**kw)
        cfgs.append(cfg)

    return cfgs


##


def _main() -> int:
    parser = argparse.ArgumentParser()

    _add_arguments(parser)

    parser.add_argument('-m', '--module', action='store_true')
    parser.add_argument('target')
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()
    cfgs = _process_arguments(args)

    with bootstrap(*cfgs):
        tgt = args.target

        sys.argv = [tgt, *(args.args or ())]

        if args.module:
            runpy._run_module_as_main(tgt)  # type: ignore  # noqa

        else:
            with io.open_code(tgt) as fp:
                src = fp.read()

            ns = dict(
                __name__='__main__',
                __file__=tgt,
                __builtins__=__builtins__,
                __spec__=None,
            )

            import __main__  # noqa
            __main__.__dict__.clear()
            __main__.__dict__.update(ns)
            exec(compile(src, tgt, 'exec'), __main__.__dict__, __main__.__dict__)

        return 0


if __name__ == '__main__':
    sys.exit(_main())



import inspect
import json
import math

from . import exceptions


# python types -> jmespath types
TYPES_MAP = {
    'bool': 'boolean',
    'list': 'array',
    'dict': 'object',
    'NoneType': 'null',
    'unicode': 'string',
    'str': 'string',
    'float': 'number',
    'int': 'number',
    'long': 'number',
    'OrderedDict': 'object',
    '_Projection': 'array',
    '_Expression': 'expref',
}


# jmespath types -> python types
REVERSE_TYPES_MAP = {
    'boolean': ('bool',),
    'array': ('list', '_Projection'),
    'object': ('dict', 'OrderedDict'),
    'null': ('NoneType',),
    'string': ('unicode', 'str'),
    'number': ('float', 'int', 'long'),
    'expref': ('_Expression',),
}


def signature(*arguments):
    def _record_signature(func):
        func.signature = arguments
        return func
    return _record_signature


class FunctionRegistry(type):
    def __init__(cls, name, bases, attrs):
        cls._populate_function_table()
        super().__init__(name, bases, attrs)

    def _populate_function_table(cls):
        function_table = {}

        # Any method with a @signature decorator that also starts with "_func_" is registered as a function.
        # _func_max_by -> max_by function.
        for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            if not name.startswith('_func_'):
                continue

            signature = getattr(method, 'signature', None)
            if signature is not None:
                function_table[name[6:]] = {
                    'function': method,
                    'signature': signature,
                }

        cls.FUNCTION_TABLE = function_table


class Functions(metaclass=FunctionRegistry):

    FUNCTION_TABLE: dict = {}  # noqa

    def call_function(self, function_name, resolved_args):
        try:
            spec = self.FUNCTION_TABLE[function_name]
        except KeyError:
            raise exceptions.UnknownFunctionError(f'Unknown function: {function_name}()')  # noqa

        function = spec['function']
        signature = spec['signature']

        self._validate_arguments(resolved_args, signature, function_name)

        return function(self, *resolved_args)

    def _validate_arguments(self, args, signature, function_name):

        if len(signature) == 0:
            return self._type_check(args, signature, function_name)

        required_arguments_count = len([
            param for param in signature if param and (not param.get('optional') or not param['optional'])
        ])
        optional_arguments_count = len([
            param for param in signature if param and param.get('optional') and param['optional']
        ])
        has_variadic = signature[-1].get('variadic') if signature is not None else False

        if has_variadic:
            if len(args) < len(signature):
                raise exceptions.VariadicArityError(len(signature), len(args), function_name)

        elif optional_arguments_count > 0:
            if (
                    len(args) < required_arguments_count or
                    len(args) > (required_arguments_count + optional_arguments_count)
            ):
                raise exceptions.ArityError(len(signature), len(args), function_name)
        elif len(args) != required_arguments_count:
            raise exceptions.ArityError(len(signature), len(args), function_name)

        return self._type_check(args, signature, function_name)

    def _type_check(self, actual, signature, function_name):
        for i in range(min(len(signature), len(actual))):
            allowed_types = self._get_allowed_types_from_signature(signature[i])
            if allowed_types:
                self._type_check_single(actual[i], allowed_types, function_name)

    def _type_check_single(self, current, types, function_name):
        # Type checking involves checking the top level type, and in the case of arrays, potentially checking the types
        # of each element.
        allowed_types, allowed_subtypes = self._get_allowed_pytypes(types)

        # We're not using isinstance() on purpose. The type model for jmespath does not map 1-1 with python types
        # (booleans are considered integers in python for example).
        actual_typename = type(current).__name__
        if actual_typename not in allowed_types:
            raise exceptions.JmespathTypeError(
                function_name, current,
                self._convert_to_jmespath_type(actual_typename), types)

        # If we're dealing with a list type, we can have additional restrictions on the type of the list elements (for
        # example a function can require a list of numbers or a list of strings). Arrays are the only types that can
        # have subtypes.
        if allowed_subtypes:
            self._subtype_check(current, allowed_subtypes, types, function_name)

    def _get_allowed_types_from_signature(self, spec):
        # signature supports monotype {'type': 'type-name'}## or multiple types {'types': ['type1-name', 'type2-name']}
        if spec.get('type'):
            spec.update({'types': [spec.get('type')]})
        return spec.get('types')

    def _get_allowed_pytypes(self, types):
        allowed_types: list = []
        allowed_subtypes: list = []

        for t in types:
            type_ = t.split('-', 1)
            if len(type_) == 2:
                type_, subtype = type_
                allowed_subtypes.append(REVERSE_TYPES_MAP[subtype])
            else:
                type_ = type_[0]

            allowed_types.extend(REVERSE_TYPES_MAP[type_])

        return allowed_types, allowed_subtypes

    def _subtype_check(self, current, allowed_subtypes, types, function_name):
        if len(allowed_subtypes) == 1:
            # The easy case, we know up front what type we need to validate.
            allowed_subtypes = allowed_subtypes[0]
            for element in current:
                actual_typename = type(element).__name__
                if actual_typename not in allowed_subtypes:
                    raise exceptions.JmespathTypeError(function_name, element, actual_typename, types)

        elif len(allowed_subtypes) > 1 and current:
            # Dynamic type validation.  Based on the first type we see, we validate that the remaining types match.
            first = type(current[0]).__name__
            for subtypes in allowed_subtypes:
                if first in subtypes:
                    allowed = subtypes
                    break
            else:
                raise exceptions.JmespathTypeError(function_name, current[0], first, types)

            for element in current:
                actual_typename = type(element).__name__
                if actual_typename not in allowed:
                    raise exceptions.JmespathTypeError(function_name, element, actual_typename, types)

    @signature({'types': ['number']})
    def _func_abs(self, arg):
        return abs(arg)

    @signature({'types': ['string']})
    def _func_lower(self, arg):
        return arg.lower()

    @signature({'types': ['string']})
    def _func_upper(self, arg):
        return arg.upper()

    @signature({'types': ['array-number']})
    def _func_avg(self, arg):
        if arg:
            return sum(arg) / len(arg)
        else:
            return None

    @signature({'types': [], 'variadic': True})
    def _func_not_null(self, *arguments):
        for argument in arguments:
            if argument is not None:
                return argument
        return None

    @signature({'types': []})
    def _func_to_array(self, arg):
        if isinstance(arg, list):
            return arg
        else:
            return [arg]

    @signature({'types': []})
    def _func_to_string(self, arg):
        if isinstance(arg, str):
            return arg
        else:
            return json.dumps(arg, separators=(',', ':'), default=str)

    @signature({'types': []})
    def _func_to_number(self, arg):
        if isinstance(arg, (list, dict, bool)):
            return None

        elif arg is None:
            return None

        elif isinstance(arg, (int, float)):
            return arg

        else:
            try:
                return int(arg)
            except ValueError:
                try:
                    return float(arg)
                except ValueError:
                    return None

    @signature({'types': ['array', 'string']}, {'types': []})
    def _func_contains(self, subject, search):
        return search in subject

    @signature({'types': ['string', 'array', 'object']})
    def _func_length(self, arg):
        return len(arg)

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_ends_with(self, search, suffix):
        return search.endswith(suffix)

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_starts_with(self, search, suffix):
        return search.startswith(suffix)

    @signature({'types': ['array', 'string']})
    def _func_reverse(self, arg):
        if isinstance(arg, str):
            return arg[::-1]
        else:
            return list(reversed(arg))

    @signature({'types': ['number']})
    def _func_ceil(self, arg):
        return math.ceil(arg)

    @signature({'types': ['number']})
    def _func_floor(self, arg):
        return math.floor(arg)

    @signature({'types': ['string']}, {'types': ['array-string']})
    def _func_join(self, separator, array):
        return separator.join(array)

    @signature({'types': ['expref']}, {'types': ['array']})
    def _func_map(self, expref, arg):
        result = []
        for element in arg:
            result.append(expref.visit(expref.expression, element))
        return result

    @signature({'types': ['array-number', 'array-string']})
    def _func_max(self, arg):
        if arg:
            return max(arg)
        else:
            return None

    @signature({'types': ['object'], 'variadic': True})
    def _func_merge(self, *arguments):
        merged = {}
        for arg in arguments:
            merged.update(arg)
        return merged

    @signature({'types': ['array-number', 'array-string']})
    def _func_min(self, arg):
        if arg:
            return min(arg)
        else:
            return None

    @signature({'types': ['array-string', 'array-number']})
    def _func_sort(self, arg):
        return sorted(arg)

    @signature({'types': ['array-number']})
    def _func_sum(self, arg):
        return sum(arg)

    @signature({'types': ['object']})
    def _func_items(self, arg):
        return [list(t) for t in arg.items()]

    @signature({'types': ['array']})
    def _func_from_items(self, items):
        return dict(items)

    @signature({'types': ['object']})
    def _func_keys(self, arg):
        # To be consistent with .values() should we also return the indices of a list?
        return list(arg.keys())

    @signature({'types': ['object']})
    def _func_values(self, arg):
        return list(arg.values())

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
        {'type': 'number', 'optional': True},
    )
    def _func_find_first(self, text, search, start=0, end=None):
        self._ensure_integer('find_first', 'start', start)
        self._ensure_integer('find_first', 'end', end)
        return self._find_impl(
            text,
            search,
            lambda t, s: t.find(s),
            start,
            end,
        )

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
        {'type': 'number', 'optional': True})
    def _func_find_last(self, text, search, start=0, end=None):
        self._ensure_integer('find_last', 'start', start)
        self._ensure_integer('find_last', 'end', end)
        return self._find_impl(
            text,
            search,
            lambda t, s: t.rfind(s),
            start,
            end,
        )

    def _find_impl(self, text, search, func, start, end):
        if len(search) == 0:
            return None
        if end is None:
            end = len(text)

        pos = func(text[start:end], search)
        if start < 0:
            start = start + len(text)

        # restrict resulting range to valid indices
        start = min(max(start, 0), len(text))
        return start + pos if pos != -1 else None

    @signature(
        {'type': 'string'},
        {'type': 'number'},
        {'type': 'string', 'optional': True},
    )
    def _func_pad_left(self, text, width, padding=' '):
        self._ensure_non_negative_integer('pad_left', 'width', width)
        return self._pad_impl(lambda: text.rjust(width, padding), padding)

    @signature(
        {'type': 'string'},
        {'type': 'number'},
        {'type': 'string', 'optional': True},
    )
    def _func_pad_right(self, text, width, padding=' '):
        self._ensure_non_negative_integer('pad_right', 'width', width)
        return self._pad_impl(lambda: text.ljust(width, padding), padding)

    def _pad_impl(self, func, padding):
        if len(padding) != 1:
            raise exceptions.JmespathError(
                f'syntax-error: pad_right() expects $padding to have a single character, but received '
                f'`{padding}` instead.',
            )
        return func()

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
    )
    def _func_replace(self, text, search, replacement, count=None):
        self._ensure_non_negative_integer(
            'replace',
            'count',
            count,
        )

        if count is not None:
            return text.replace(search, replacement, int(count))

        return text.replace(search, replacement)

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
    )
    def _func_split(self, text, search, count=None):
        self._ensure_non_negative_integer(
            'split',
            'count',
            count,
        )

        if len(search) == 0:
            chars = list(text)
            if count is None:
                return chars

            head = list(chars[:count])
            tail = [''.join(chars[count:])]
            return head + tail

        if count is not None:
            return text.split(search, count)

        return text.split(search)

    def _ensure_integer(
            self,
            func_name,
            param_name,
            param_value,
    ):
        if param_value is not None:
            if int(param_value) != param_value:
                raise exceptions.JmespathValueError(
                    func_name,
                    param_value,
                    'integer',
                )

    def _ensure_non_negative_integer(
            self,
            func_name,
            param_name,
            param_value,
    ):
        if param_value is not None:
            if int(param_value) != param_value or int(param_value) < 0:
                raise exceptions.JmespathValueError(
                    func_name,
                    param_name,
                    'non-negative integer',
                )

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.strip()
        return text.strip(chars)

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim_left(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.lstrip()
        return text.lstrip(chars)

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim_right(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.rstrip()
        return text.rstrip(chars)

    @signature({'types': []})
    def _func_type(self, arg):
        if isinstance(arg, str):
            return 'string'
        elif isinstance(arg, bool):
            return 'boolean'
        elif isinstance(arg, list):
            return 'array'
        elif isinstance(arg, dict):
            return 'object'
        elif isinstance(arg, (float, int)):
            return 'number'
        elif arg is None:
            return 'null'
        else:
            return None

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_sort_by(self, array, expref):
        if not array:
            return array

        # sort_by allows for the expref to be either a number of a string, so we have some special logic to handle this.
        # We evaluate the first array element and verify that it's either a string of a number.  We then create a key
        # function that validates that type, which requires that remaining array elements resolve to the same type as
        # the first element.
        required_type = self._convert_to_jmespath_type(type(expref.visit(expref.expression, array[0])).__name__)
        if required_type not in ['number', 'string']:
            raise exceptions.JmespathTypeError(
                'sort_by',
                array[0],
                required_type,
                ['string', 'number'],
            )

        keyfunc = self._create_key_func(expref, [required_type], 'sort_by')
        return sorted(array, key=keyfunc)

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_min_by(self, array, expref):
        keyfunc = self._create_key_func(
            expref,
            ['number', 'string'],
            'min_by',
        )

        if array:
            return min(array, key=keyfunc)
        else:
            return None

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_max_by(self, array, expref):
        keyfunc = self._create_key_func(
            expref,
            ['number', 'string'],
            'max_by',
        )

        if array:
            return max(array, key=keyfunc)
        else:
            return None

    @signature({'types': ['array'], 'variadic': True})
    def _func_zip(self, *arguments):
        return [list(t) for t in zip(*arguments)]

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_group_by(self, array, expref):
        keyfunc = self._create_key_func(expref, ['null', 'string'], 'group_by')
        if not array:
            return None

        result = {}
        keys = list(dict.fromkeys([keyfunc(item) for item in array if keyfunc(item) is not None]))
        for key in keys:
            items = [item for item in array if keyfunc(item) == key]
            result.update({key: items})

        return result

    def _create_key_func(self, expref, allowed_types, function_name):
        def keyfunc(x):
            result = expref.visit(expref.expression, x)
            actual_typename = type(result).__name__

            jmespath_type = self._convert_to_jmespath_type(actual_typename)
            # allowed_types is in terms of jmespath types, not python types.
            if jmespath_type not in allowed_types:
                raise exceptions.JmespathTypeError(
                    function_name, result, jmespath_type, allowed_types)

            return result

        return keyfunc

    def _convert_to_jmespath_type(self, pyobject):
        return TYPES_MAP.get(pyobject, 'unknown')



import collections
import dataclasses as dc
import functools
import heapq
import itertools
import typing as ta

# from . import check
from . import lang


T = ta.TypeVar('T')
U = ta.TypeVar('U')

_MISSING = object()


class PeekIterator(ta.Iterator[T]):

    def __init__(self, it: ta.Iterable[T]) -> None:
        super().__init__()

        self._it = iter(it)
        self._pos = -1
        self._next_item: ta.Any = _MISSING

    _item: T

    def __iter__(self) -> ta.Self:
        return self

    @property
    def done(self) -> bool:
        try:
            self.peek()
        except StopIteration:
            return True
        else:
            return False

    def __next__(self) -> T:
        if self._next_item is not _MISSING:
            self._item = ta.cast(T, self._next_item)
            self._next_item = _MISSING
        else:
            self._item = next(self._it)
        self._pos += 1
        return self._item

    def peek(self) -> T:
        if self._next_item is not _MISSING:
            return ta.cast(T, self._next_item)
        self._next_item = next(self._it)
        return self._next_item

    def next_peek(self) -> T:
        next(self)
        return self.peek()

    def takewhile(self, fn: ta.Callable[[T], bool]) -> ta.Iterator[T]:
        while fn(self.peek()):
            yield next(self)

    def skipwhile(self, fn: ta.Callable[[T], bool]) -> None:
        while fn(self.peek()):
            next(self)

    def takeuntil(self, fn: ta.Callable[[T], bool]) -> ta.Iterator[T]:
        return self.takewhile(lambda e: not fn(e))

    def skipuntil(self, fn: ta.Callable[[T], bool]) -> None:
        self.skipwhile(lambda e: not fn(e))

    def takethrough(self, pos: int) -> ta.Iterator[T]:
        return self.takewhile(lambda _: self._pos < pos)

    def skipthrough(self, pos: int) -> None:
        self.skipwhile(lambda _: self._pos < pos)

    def taketo(self, pos: int) -> ta.Iterator[T]:
        return self.takethrough(pos - 1)

    def skipto(self, pos: int) -> None:
        self.skipthrough(pos - 1)


class ProxyIterator(ta.Iterator[T]):

    def __init__(self, fn: ta.Callable[[], T]) -> None:
        self._fn = fn

    def __iter__(self) -> ta.Self:
        return self

    def __next__(self) -> T:
        return self._fn()


class PrefetchIterator(ta.Iterator[T]):

    def __init__(self, fn: ta.Callable[[], T] | None = None) -> None:
        super().__init__()

        self._fn = fn
        self._deque: collections.deque[T] = collections.deque()

    def __iter__(self) -> ta.Self:
        return self

    def push(self, item) -> None:
        self._deque.append(item)

    def __next__(self) -> T:
        try:
            return self._deque.popleft()
        except IndexError:
            if self._fn is None:
                raise StopIteration from None
        return self._fn()


class RetainIterator(ta.Iterator[T]):

    def __init__(self, fn: ta.Callable[[], T]) -> None:
        super().__init__()

        self._fn = fn
        self._deque: collections.deque[T] = collections.deque()

    def __iter__(self) -> ta.Self:
        return self

    def pop(self) -> None:
        self._deque.popleft()

    def __next__(self) -> T:
        item = self._fn()
        self._deque.append(item)
        return item


def unzip(it: ta.Iterable[T], width: int | None = None) -> list:
    if width is None:
        if not isinstance(it, PeekIterator):
            it = PeekIterator(iter(it))
        try:
            width = len(it.peek())
        except StopIteration:
            return []

    its: list[PrefetchIterator[T]] = []
    running = True

    def next_fn(idx):
        nonlocal running
        if not running:
            raise StopIteration
        try:
            items = next(it)  # type: ignore
        except StopIteration:
            running = False
            raise
        for item_idx, item in enumerate(items):
            its[item_idx].push(item)
        return next(its[idx])

    its.extend(PrefetchIterator(functools.partial(next_fn, idx)) for idx in range(width))
    return its


def take(n: int, iterable: ta.Iterable[T]) -> list[T]:
    return list(itertools.islice(iterable, n))


def chunk(n: int, iterable: ta.Iterable[T], strict: bool = False) -> ta.Iterator[list[T]]:
    iterator = iter(functools.partial(take, n, iter(iterable)), [])
    if strict:
        def ret():
            for chunk in iterator:
                if len(chunk) != n:
                    raise ValueError('iterable is not divisible by n.')
                yield chunk
        return iter(ret())
    else:
        return iterator


def merge_on(
        function: ta.Callable[[T], U],
        *its: ta.Iterable[T],
) -> ta.Iterator[tuple[U, list[tuple[int, T]]]]:
    indexed_its = [
        (
            (function(item), it_idx, item)
            for it_idx, item in zip(itertools.repeat(it_idx), it)
        )
        for it_idx, it in enumerate(its)
    ]

    grouped_indexed_its = itertools.groupby(
        heapq.merge(*indexed_its),
        key=lambda item_tuple: item_tuple[0],
    )

    return (
        (fn_item, [(it_idx, item) for _, it_idx, item in grp])
        for fn_item, grp in grouped_indexed_its
    )


def expand_indexed_pairs(
        seq: ta.Iterable[tuple[int, T]],
        default: T,
        *,
        width: int | None = None,
) -> list[T]:
    width_ = width
    if width_ is None:
        width_ = (max(idx for idx, _ in seq) + 1) if seq else 0
    result = [default] * width_
    for idx, value in seq:
        if idx < width_:
            result[idx] = value
    return result


##
# https://docs.python.org/3/library/itertools.html#itertools-recipes


def sliding_window(it: ta.Iterable[T], n: int) -> ta.Iterator[tuple[T, ...]]:
    # sliding_window('ABCDEFG', 4) -> ABCD BCDE CDEF DEFG
    iterator = iter(it)
    window = collections.deque(itertools.islice(iterator, n - 1), maxlen=n)
    for x in iterator:
        window.append(x)
        yield tuple(window)


##


@dc.dataclass()
class UniqueStats:
    key: ta.Any
    num_seen: int
    first_idx: int
    last_idx: int


@dc.dataclass(frozen=True)
class UniqueItem(ta.Generic[T]):
    idx: int
    item: T
    stats: UniqueStats
    out: lang.Maybe[T]


class UniqueIterator(ta.Iterator[UniqueItem[T]]):
    def __init__(
            self,
            it: ta.Iterable[T],
            keyer: ta.Callable[[T], ta.Any] = lang.identity,
    ) -> None:
        super().__init__()
        self._it = enumerate(it)
        self._keyer = keyer

        self.stats: dict[ta.Any, UniqueStats] = {}

    def __next__(self) -> UniqueItem[T]:
        idx, item = next(self._it)
        key = self._keyer(item)

        try:
            stats = self.stats[key]

        except KeyError:
            stats = self.stats[key] = UniqueStats(
                key,
                num_seen=1,
                first_idx=idx,
                last_idx=idx,
            )
            return UniqueItem(
                idx,
                item,
                stats,
                lang.just(item),
            )

        else:
            stats.num_seen += 1
            stats.last_idx = idx
            return UniqueItem(
                idx,
                item,
                stats,
                lang.empty(),
            )



import collections.abc
import dataclasses as dc
import functools
import typing as ta

from .. import check
from .. import matchfns as mfs
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactoryMatchClass
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactoryMatchClass
from .values import Value


DEFAULT_ITERABLE_CONCRETE_TYPES: dict[type[collections.abc.Iterable], type[collections.abc.Iterable]] = {
    collections.abc.Iterable: tuple,  # type: ignore
    collections.abc.Sequence: tuple,  # type: ignore
    collections.abc.MutableSequence: list,  # type: ignore
}


@dc.dataclass(frozen=True)
class IterableMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Iterable) -> Value:
        return list(map(functools.partial(self.e.marshal, ctx), o))


class IterableMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable))
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return IterableMarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Iterable))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return IterableMarshaler(ctx.make(ta.Any))


@dc.dataclass(frozen=True)
class IterableUnmarshaler(Unmarshaler):
    ctor: ta.Callable[[ta.Iterable[ta.Any]], ta.Iterable]
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Iterable:
        return self.ctor(map(functools.partial(self.e.unmarshal, ctx), check.isinstance(v, collections.abc.Iterable)))


class IterableUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable))
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        cty = DEFAULT_ITERABLE_CONCRETE_TYPES.get(gty.cls, gty.cls)  # noqa
        return IterableUnmarshaler(cty, ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Iterable))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return IterableUnmarshaler(check.isinstance(rty, type), ctx.make(ta.Any))



import typing as ta


V = ta.TypeVar('V')


class HttpMap(ta.Mapping[str, V]):
    def __getitem__(self, k):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError



import collections.abc
import typing as ta
import weakref


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


def multikey_dict(dct: ta.Mapping[ta.Iterable[K] | K, V], *, deep: bool = False) -> dict[K, V]:
    ret = {}
    for k, v in dct.items():
        if deep and isinstance(v, dict):
            v = multikey_dict(v, deep=True)  # type: ignore
        if isinstance(k, tuple):
            for sk in k:
                ret[sk] = v
        else:
            ret[k] = v
    return ret


def guarded_map_update(
        dst: ta.MutableMapping[ta.Any, ta.Any],
        *srcs: ta.Mapping[ta.Any, ta.Any],
) -> ta.MutableMapping[ta.Any, ta.Any]:
    for src in srcs:
        for k, v in src.items():
            if k in dst:
                raise KeyError(k)
            dst[k] = v
    return dst


def yield_dict_init(*args, **kwargs) -> ta.Iterable[tuple[ta.Any, ta.Any]]:
    if len(args) > 1:
        raise TypeError
    if args:
        [src] = args
        if isinstance(src, collections.abc.Mapping):
            for k in src:
                yield (k, src[k])
        else:
            for k, v in src:
                yield (k, v)
    for k, v in kwargs.items():
        yield (k, v)


class TypeMap(ta.Generic[T]):

    def __init__(self, items: ta.Iterable[T] = ()) -> None:
        super().__init__()

        self._items = list(items)
        dct: dict[type, ta.Any] = {}
        for item in items:
            if (ty := type(item)) in dct:
                raise ValueError(ty)
            dct[ty] = item
        self._dct = dct

    @property
    def items(self) -> ta.Sequence[T]:
        return self._items

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._items)

    def get(self, ty: type[T]) -> T | None:
        return self._dct.get(ty)

    def __getitem__(self, ty: type[T]) -> ta.Sequence[T]:
        return self._dct[ty]


class DynamicTypeMap(ta.Generic[V]):

    def __init__(self, items: ta.Iterable[V] = (), *, weak: bool = False) -> None:
        super().__init__()

        self._items = list(items)
        self._weak = bool(weak)

        self._cache: ta.MutableMapping[type, ta.Any] = weakref.WeakKeyDictionary()

    @property
    def items(self) -> ta.Sequence[V]:
        return self._items

    @property
    def weak(self) -> bool:
        return self._weak

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> ta.Iterator[V]:
        return iter(self._items)

    def __getitem__(self, ty: type[T]) -> ta.Sequence[T]:
        try:
            return self._cache[ty]
        except KeyError:
            ret = []
            for item in self._items:
                if isinstance(item, ty):
                    ret.append(item)
            self._cache[ty] = ret
            return ret


class MissingDict(dict[K, V]):

    def __init__(self, missing_fn: ta.Callable[[K], V]) -> None:
        if not callable(missing_fn):
            raise TypeError(missing_fn)
        super().__init__()
        self._missing_fn = missing_fn

    def __missing__(self, key):
        v = self[key] = self._missing_fn(key)
        return v



import base64
import dataclasses as dc
import datetime
import hashlib
import hmac
import struct
import time
import typing as ta
import zlib

from .. import fnpairs as fpa
from .. import lang
from .. import secrets as sec
from .cookies import dump_cookie
from .cookies import parse_cookie
from .json import JSON_TAGGER


Session: ta.TypeAlias = dict[str, ta.Any]


##


def base64_encode(b: bytes) -> bytes:
    return base64.urlsafe_b64encode(b).rstrip(b'=')


def base64_decode(b: bytes) -> bytes:
    b += b'=' * (-len(b) % 4)
    return base64.urlsafe_b64decode(b)


def int_to_bytes(num: int) -> bytes:
    return struct.pack('>Q', num).lstrip(b'\0')


def bytes_to_int(bytestr: bytes) -> int:
    return struct.unpack('>Q', bytestr.rjust(8, b'\0'))[0]


##


class Signer:
    @dc.dataclass(frozen=True)
    class Config:
        secret_key: str | sec.SecretRef = dc.field()
        salt: str = 'cookie-session'

    def __init__(
            self,
            config: Config,
            *,
            secrets: sec.Secrets = sec.EMPTY_SECRETS,
    ) -> None:
        super().__init__()

        self._config = config
        self._secrets = secrets

    @lang.cached_function
    def digest(self) -> ta.Any:
        return hashlib.sha1

    @lang.cached_function
    def derive_key(self) -> bytes:
        mac = hmac.new(self._secrets.fix(self._config.secret_key).reveal().encode(), digestmod=self.digest())
        mac.update(self._config.salt.encode())
        return mac.digest()

    def get_signature(self, value: bytes) -> bytes:
        mac = hmac.new(self.derive_key(), msg=value, digestmod=self.digest())
        return mac.digest()

    def verify_signature(self, value: bytes, sig: bytes) -> bool:
        return hmac.compare_digest(sig, self.get_signature(value))


##


class SessionExpiredError(Exception):
    pass


class SessionVerificationError(Exception):
    pass


class SessionMarshal:
    def __init__(
            self,
            signer: Signer,
            serializer: fpa.ObjectStr = fpa.of(JSON_TAGGER.dumps, JSON_TAGGER.loads),
    ) -> None:
        super().__init__()

        self._signer = signer
        self._serializer = serializer

    SEP = b'.'

    def load(self, bs: bytes) -> ta.Any:
        value, sig = bs.rsplit(self.SEP, 1)

        sig_b = base64_decode(sig)

        if not self._signer.verify_signature(value, sig_b):
            raise SessionVerificationError

        value, ts_bytes = value.rsplit(self.SEP, 1)
        ts_int = bytes_to_int(base64_decode(ts_bytes))

        max_age = 31 * 24 * 60 * 60
        age = int(time.time()) - ts_int

        if age > max_age:
            raise SessionExpiredError
        if age < 0:
            raise SessionExpiredError

        payload = value

        decompress = False
        if payload.startswith(b'.'):
            payload = payload[1:]
            decompress = True

        jb = base64_decode(payload)

        if decompress:
            jb = zlib.decompress(jb)

        jbs = jb.decode()

        obj = self._serializer.backward(jbs)

        return obj

    def dump(self, obj: ta.Any) -> bytes:
        jbs = self._serializer.forward(obj)

        jb = jbs.encode()

        is_compressed = False
        compressed = zlib.compress(jb)

        if len(compressed) < (len(jb) - 1):
            jb = compressed
            is_compressed = True

        base64d = base64_encode(jb)

        if is_compressed:
            base64d = b'.' + base64d

        payload = base64d

        timestamp = base64_encode(int_to_bytes(int(time.time())))

        value = payload + self.SEP + timestamp
        return value + self.SEP + base64_encode(self._signer.get_signature(value))


##


class CookieSessionStore:
    @dc.dataclass(frozen=True)
    class Config:
        key: str = 'session'
        max_age: datetime.timedelta | int | None = None

    def __init__(self, marshal: SessionMarshal, config: Config = Config()) -> None:
        super().__init__()

        self._marshal = marshal
        self._config = config

    def extract(self, scope) -> Session:
        for k, v in scope['headers']:
            if k == b'cookie':
                cks = parse_cookie(v.decode('latin-1', 'strict'))
                sk = cks.get(self._config.key)
                if sk:
                    return self._marshal.load(sk[0].encode('latin-1', 'strict'))

        return {}

    def build_headers(self, session: Session) -> list[tuple[bytes, bytes]]:
        d = self._marshal.dump(session)

        c = dump_cookie(
            self._config.key,
            d.decode('latin-1', 'strict'),
            max_age=self._config.max_age,
            httponly=True,
        )

        return [
            (b'Vary', b'Cookie'),
            (b'Set-Cookie', c.encode('latin-1', 'strict')),
        ]



import typing as ta
import urllib.parse

from .. import dataclasses as dc
from .. import lang
from .. import secrets as sec


##


@dc.dataclass(frozen=True, kw_only=True)
class DbType(lang.Final):
    name: str
    dialect_name: str

    default_port: int | None = None


class DbTypes(lang.Namespace, lang.Final):
    MYSQL = DbType(
        name='mysql',
        dialect_name='mysql',
        default_port=3306,
    )

    POSTGRES = DbType(
        name='postgres',
        dialect_name='postgresql',
        default_port=5432,
    )

    SQLITE = DbType(
        name='sqlite',
        dialect_name='sqlite',
    )


##


class DbLoc(lang.Abstract, lang.Sealed):
    pass


@dc.dataclass(frozen=True)
class UrlDbLoc(DbLoc, lang.Final):
    url: sec.SecretRefOrStr = dc.xfield() | sec.secret_field


@dc.dataclass(frozen=True)
class HostDbLoc(DbLoc, lang.Final):
    host: str
    port: int | None = None

    username: str | None = None
    password: sec.SecretRefOrStr | None = dc.xfield(default=None) | sec.secret_field


##


@dc.dataclass(frozen=True)
class DbSpec(lang.Final):
    name: str
    type: DbType
    loc: DbLoc


##


def rebuild_url(url: str, fn: ta.Callable[[urllib.parse.ParseResult], urllib.parse.ParseResult]) -> str:
    if '://' in url:
        engine, _, url = url.partition('://')
        url = 'sql://' + url
    else:
        engine = None
    parsed = urllib.parse.urlparse(url)
    parsed = fn(parsed)
    if engine is not None and parsed.scheme == 'sql':
        parsed = parsed._replace(scheme=engine)
    return urllib.parse.urlunparse(parsed)  # noqa


def set_url_engine(url: str, engine: str) -> str:
    return rebuild_url(url, lambda parsed: parsed._replace(scheme=engine))  # noqa


def set_url_database(url: str, database: str) -> str:
    return rebuild_url(url, lambda parsed: parsed._replace(path='/' + database))  # noqa
